@@ -252,8 +252,11 @@ impl Parse for BareType {
 
 pub mod kw {
     syn::custom_keyword!(name);
+    syn::custom_keyword!(span);
     syn::custom_keyword!(title);
     syn::custom_keyword!(scope);
+    syn::custom_keyword!(contextual);
+    syn::custom_keyword!(cast);
     syn::custom_keyword!(constructor);
     syn::custom_keyword!(keywords);
     syn::custom_keyword!(parent);
