@@ -17,6 +17,7 @@ pub fn module() -> Module {
     scope.define_func::<pow>();
     scope.define_func::<exp>();
     scope.define_func::<sqrt>();
+    scope.define_func::<root>();
     scope.define_func::<sin>();
     scope.define_func::<cos>();
     scope.define_func::<tan>();
@@ -113,11 +114,9 @@ pub fn pow(
     };
 
     let result = match (base, exponent.v) {
-        (Num::Int(a), Num::Int(b)) if b >= 0 => a
-            .checked_pow(b as u32)
-            .map(Num::Int)
-            .ok_or("the result is too large")
-            .at(span)?,
+        (Num::Int(a), Num::Int(b)) if b >= 0 => {
+            a.checked_pow(b as u32).map(Num::Int).ok_or_else(too_large).at(span)?
+        }
         (a, b) => Num::Float(if a.float() == std::f64::consts::E {
             b.float().exp()
         } else if a.float() == 2.0 {
@@ -183,6 +182,37 @@ pub fn sqrt(
     Ok(value.v.float().sqrt())
 }
 
+/// Calculates the real nth root of a number.
+///
+/// If the number is negative, then n must be odd.
+///
+/// ```example
+/// #calc.root(16.0, 4) \
+/// #calc.root(27.0, 3)
+/// ```
+#[func]
+pub fn root(
+    /// The expression to take the root of
+    radicand: f64,
+    /// Which root of the radicand to take
+    index: Spanned<i64>,
+) -> SourceResult<f64> {
+    if index.v == 0 {
+        bail!(index.span, "cannot take the 0th root of a number");
+    } else if radicand < 0.0 {
+        if index.v % 2 == 0 {
+            bail!(
+                index.span,
+                "negative numbers do not have a real nth root when n is even"
+            );
+        } else {
+            Ok(-(-radicand).powf(1.0 / index.v as f64))
+        }
+    } else {
+        Ok(radicand.powf(1.0 / index.v as f64))
+    }
+}
+
 /// Calculates the sine of an angle.
 ///
 /// When called with an integer or a float, they will be interpreted as
@@ -437,7 +467,7 @@ pub fn fact(
     /// The number whose factorial to calculate. Must be non-negative.
     number: u64,
 ) -> StrResult<i64> {
-    Ok(fact_impl(1, number).ok_or("the result is too large")?)
+    Ok(fact_impl(1, number).ok_or_else(too_large)?)
 }
 
 /// Calculates a permutation.
@@ -461,7 +491,7 @@ pub fn perm(
         return Ok(0);
     }
 
-    Ok(fact_impl(base - numbers + 1, base).ok_or("the result is too large")?)
+    Ok(fact_impl(base - numbers + 1, base).ok_or_else(too_large)?)
 }
 
 /// Calculates the product of a range of numbers. Used to calculate
@@ -496,7 +526,7 @@ pub fn binom(
     /// The lower coefficient. Must be non-negative.
     k: u64,
 ) -> StrResult<i64> {
-    Ok(binom_impl(n, k).ok_or("the result is too large")?)
+    Ok(binom_impl(n, k).ok_or_else(too_large)?)
 }
 
 /// Calculates a binomial coefficient, with `n` the upper coefficient and `k`
@@ -562,7 +592,7 @@ pub fn lcm(
     Ok(a.checked_div(gcd(a, b))
         .and_then(|gcd| gcd.checked_mul(b))
         .map(|v| v.abs())
-        .ok_or("the return value is too large")?)
+        .ok_or_else(too_large)?)
 }
 
 /// Rounds a number down to the nearest integer.
@@ -942,3 +972,9 @@ cast! {
     v: f64 => Self::Float(v),
     v: Angle => Self::Angle(v),
 }
+
+/// The error message when the result is too large to be represented.
+#[cold]
+fn too_large() -> &'static str {
+    "the result is too large"
+}
