@@ -3,6 +3,7 @@ use std::fmt::{self, Debug, Display, Formatter};
 use std::hash::{Hash, Hasher};
 use std::ops::{Add, AddAssign, Deref, Range};
 
+use comemo::Tracked;
 use ecow::EcoString;
 use serde::{Deserialize, Serialize};
 use unicode_segmentation::UnicodeSegmentation;
@@ -10,10 +11,10 @@ use unicode_segmentation::UnicodeSegmentation;
 use crate::diag::{bail, At, SourceResult, StrResult};
 use crate::engine::Engine;
 use crate::foundations::{
-    cast, dict, func, repr, scope, ty, Array, Bytes, Dict, Func, IntoValue, Label, Repr,
-    Type, Value, Version,
+    cast, dict, func, repr, scope, ty, Array, Bytes, Context, Dict, Func, IntoValue,
+    Label, Repr, Type, Value, Version,
 };
-use crate::layout::Align;
+use crate::layout::Alignment;
 use crate::syntax::{Span, Spanned};
 
 /// Create a new [`Str`] from a format string.
@@ -67,7 +68,7 @@ pub use ecow::eco_format;
 /// - `[\r]` for a carriage return
 /// - `[\t]` for a tab
 /// - `[\u{1f600}]` for a hexadecimal Unicode escape sequence
-#[ty(scope, title = "String")]
+#[ty(scope, cast, title = "String")]
 #[derive(Default, Clone, Eq, PartialEq, Ord, PartialOrd, Hash)]
 #[derive(Serialize, Deserialize)]
 #[serde(transparent)]
@@ -114,7 +115,7 @@ impl Str {
             .and_then(|v| usize::try_from(v).ok())
             .filter(|&v| v <= self.0.len());
 
-        if resolved.map_or(false, |i| !self.0.is_char_boundary(i)) {
+        if resolved.is_some_and(|i| !self.0.is_char_boundary(i)) {
             return Err(not_a_char_boundary(index));
         }
 
@@ -308,7 +309,7 @@ impl Str {
     ) -> bool {
         match pattern {
             StrPattern::Str(pat) => self.0.starts_with(pat.as_str()),
-            StrPattern::Regex(re) => re.find(self).map_or(false, |m| m.start() == 0),
+            StrPattern::Regex(re) => re.find(self).is_some_and(|m| m.start() == 0),
         }
     }
 
@@ -424,6 +425,8 @@ impl Str {
         &self,
         /// The engine.
         engine: &mut Engine,
+        /// The callsite context.
+        context: Tracked<Context>,
         /// The pattern to search for.
         pattern: StrPattern,
         /// The string to replace the matches with or a function that gets a
@@ -449,8 +452,10 @@ impl Str {
             match &replacement {
                 Replacement::Str(s) => output.push_str(s),
                 Replacement::Func(func) => {
-                    let piece =
-                        func.call(engine, [dict])?.cast::<Str>().at(func.span())?;
+                    let piece = func
+                        .call(engine, context, [dict])?
+                        .cast::<Str>()
+                        .at(func.span())?;
                     output.push_str(&piece);
                 }
             }
@@ -762,7 +767,7 @@ pub enum ToStr {
 cast! {
     ToStr,
     v: i64 => Self::Int(v),
-    v: f64 => Self::Str(repr::format_float(v, None, "").into()),
+    v: f64 => Self::Str(repr::display_float(v).into()),
     v: Version => Self::Str(format_str!("{}", v)),
     v: Bytes => Self::Str(
         std::str::from_utf8(&v)
@@ -898,10 +903,6 @@ impl Hash for Regex {
     }
 }
 
-cast! {
-    type Regex,
-}
-
 /// A pattern which can be searched for in a string.
 #[derive(Debug, Clone)]
 pub enum StrPattern {
@@ -933,9 +934,9 @@ pub enum StrSide {
 
 cast! {
     StrSide,
-    v: Align => match v {
-        Align::START => Self::Start,
-        Align::END => Self::End,
+    v: Alignment => match v {
+        Alignment::START => Self::Start,
+        Alignment::END => Self::End,
         _ => bail!("expected either `start` or `end`"),
     },
 }
