@@ -1,28 +1,26 @@
 use std::any::{Any, TypeId};
-use std::borrow::Cow;
 use std::fmt::{self, Debug, Formatter};
 use std::hash::{Hash, Hasher};
-use std::{iter, mem, ptr};
+use std::{mem, ptr};
 
-use comemo::Prehashed;
+use comemo::{Track, Tracked};
 use ecow::{eco_vec, EcoString, EcoVec};
-use once_cell::sync::Lazy;
 use smallvec::SmallVec;
 
 use crate::diag::{SourceResult, Trace, Tracepoint};
 use crate::engine::Engine;
 use crate::foundations::{
-    cast, elem, func, ty, Content, Element, Func, NativeElement, Repr, Selector, Show,
+    cast, elem, func, ty, Content, Context, Element, Func, NativeElement, Packed, Repr,
+    Selector, Show,
 };
+use crate::introspection::Locatable;
 use crate::syntax::Span;
 use crate::text::{FontFamily, FontList, TextElem};
+use crate::util::LazyHash;
 
 /// Provides access to active styles.
 ///
-/// The styles are currently opaque and only useful in combination with the
-/// [`measure`]($measure) function. See its documentation for more details. In
-/// the future, the provided styles might also be directly accessed to look up
-/// styles defined by [set rules]($styling/#set-rules).
+/// **Deprecation planned.** Use [context] instead.
 ///
 /// ```example
 /// #let thing(body) = style(styles => {
@@ -35,6 +33,8 @@ use crate::text::{FontFamily, FontList, TextElem};
 /// ```
 #[func]
 pub fn style(
+    /// The call site span.
+    span: Span,
     /// A function to call with the styles. Its return value is displayed
     /// in the document.
     ///
@@ -43,28 +43,32 @@ pub fn style(
     /// content that depends on the style context it appears in.
     func: Func,
 ) -> Content {
-    StyleElem::new(func).pack()
+    StyleElem::new(func).pack().spanned(span)
 }
 
 /// Executes a style access.
-#[elem(Show)]
+#[elem(Locatable, Show)]
 struct StyleElem {
     /// The function to call with the styles.
     #[required]
     func: Func,
 }
 
-impl Show for StyleElem {
-    #[tracing::instrument(name = "StyleElem::show", skip_all)]
+impl Show for Packed<StyleElem> {
+    #[typst_macros::time(name = "style", span = self.span())]
     fn show(&self, engine: &mut Engine, styles: StyleChain) -> SourceResult<Content> {
-        Ok(self.func().call(engine, [styles.to_map()])?.display())
+        let context = Context::new(self.location(), Some(styles));
+        Ok(self
+            .func()
+            .call(engine, context.track(), [styles.to_map()])?
+            .display())
     }
 }
 
 /// A list of style properties.
-#[ty]
+#[ty(cast)]
 #[derive(Default, PartialEq, Clone, Hash)]
-pub struct Styles(EcoVec<Prehashed<Style>>);
+pub struct Styles(EcoVec<LazyHash<Style>>);
 
 impl Styles {
     /// Create a new, empty style list.
@@ -77,13 +81,18 @@ impl Styles {
         self.0.is_empty()
     }
 
+    /// Iterate over the contained styles.
+    pub fn iter(&self) -> impl Iterator<Item = &Style> {
+        self.0.iter().map(|style| &**style)
+    }
+
     /// Set an inner value for a style property.
     ///
     /// If the property needs folding and the value is already contained in the
     /// style map, `self` contributes the outer values and `value` is the inner
     /// one.
     pub fn set(&mut self, style: impl Into<Style>) {
-        self.0.push(Prehashed::new(style.into()));
+        self.0.push(LazyHash::new(style.into()));
     }
 
     /// Remove the style that was last set.
@@ -99,22 +108,20 @@ impl Styles {
 
     /// Apply one outer styles.
     pub fn apply_one(&mut self, outer: Style) {
-        self.0.insert(0, Prehashed::new(outer));
+        self.0.insert(0, LazyHash::new(outer));
     }
 
     /// Apply a slice of outer styles.
-    pub fn apply_slice(&mut self, outer: &[Prehashed<Style>]) {
+    pub fn apply_slice(&mut self, outer: &[LazyHash<Style>]) {
         self.0 = outer.iter().cloned().chain(mem::take(self).0).collect();
     }
 
     /// Add an origin span to all contained properties.
     pub fn spanned(mut self, span: Span) -> Self {
         for entry in self.0.make_mut() {
-            entry.update(|entry| {
-                if let Style::Property(property) = entry {
-                    property.span = Some(span);
-                }
-            });
+            if let Style::Property(property) = &mut **entry {
+                property.span = Some(span);
+            }
         }
         self
     }
@@ -126,6 +133,7 @@ impl Styles {
         self.0.iter().find_map(|entry| match &**entry {
             Style::Property(property) => property.is_of(elem).then_some(property.span),
             Style::Recipe(recipe) => recipe.is_of(elem).then_some(Some(recipe.span)),
+            Style::Revocation(_) => None,
         })
     }
 
@@ -140,9 +148,15 @@ impl Styles {
     }
 }
 
+impl From<LazyHash<Style>> for Styles {
+    fn from(style: LazyHash<Style>) -> Self {
+        Self(eco_vec![style])
+    }
+}
+
 impl From<Style> for Styles {
-    fn from(entry: Style) -> Self {
-        Self(eco_vec![Prehashed::new(entry)])
+    fn from(style: Style) -> Self {
+        Self(eco_vec![LazyHash::new(style)])
     }
 }
 
@@ -166,6 +180,8 @@ pub enum Style {
     Property(Property),
     /// A show rule recipe.
     Recipe(Recipe),
+    /// Disables a specific show rule recipe.
+    Revocation(RecipeIndex),
 }
 
 impl Style {
@@ -191,6 +207,7 @@ impl Debug for Style {
         match self {
             Self::Property(property) => property.fmt(f),
             Self::Recipe(recipe) => recipe.fmt(f),
+            Self::Revocation(guard) => guard.fmt(f),
         }
     }
 }
@@ -222,11 +239,17 @@ pub struct Property {
 
 impl Property {
     /// Create a new property from a key-value pair.
-    pub fn new<T>(elem: Element, id: u8, value: T) -> Self
+    pub fn new<E, T>(id: u8, value: T) -> Self
     where
+        E: NativeElement,
         T: Debug + Clone + Hash + Send + Sync + 'static,
     {
-        Self { elem, id, value: Block::new(value), span: None }
+        Self {
+            elem: E::elem(),
+            id,
+            value: Block::new(value),
+            span: None,
+        }
     }
 
     /// Whether this property is the given one.
@@ -238,6 +261,11 @@ impl Property {
     pub fn is_of(&self, elem: Element) -> bool {
         self.elem == elem
     }
+
+    /// Turn this property into prehashed style.
+    pub fn wrap(self) -> LazyHash<Style> {
+        LazyHash::new(Style::Property(self))
+    }
 }
 
 impl Debug for Property {
@@ -351,18 +379,23 @@ impl Recipe {
     }
 
     /// Whether the recipe is applicable to the target.
-    pub fn applicable(&self, target: &Content) -> bool {
+    pub fn applicable(&self, target: &Content, styles: StyleChain) -> bool {
         self.selector
             .as_ref()
-            .map_or(false, |selector| selector.matches(target))
+            .is_some_and(|selector| selector.matches(target, Some(styles)))
     }
 
     /// Apply the recipe to the given content.
-    pub fn apply(&self, engine: &mut Engine, content: Content) -> SourceResult<Content> {
+    pub fn apply(
+        &self,
+        engine: &mut Engine,
+        context: Tracked<Context>,
+        content: Content,
+    ) -> SourceResult<Content> {
         let mut content = match &self.transform {
             Transformation::Content(content) => content.clone(),
             Transformation::Func(func) => {
-                let mut result = func.call(engine, [content.clone()]);
+                let mut result = func.call(engine, context, [content.clone()]);
                 if self.selector.is_some() {
                     let point = || Tracepoint::Show(content.func().name().into());
                     result = result.trace(engine.world, point, content.span());
@@ -389,6 +422,10 @@ impl Debug for Recipe {
     }
 }
 
+/// Identifies a show rule recipe from the top of the chain.
+#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
+pub struct RecipeIndex(pub usize);
+
 /// A show rule transformation that can be applied to a match.
 #[derive(Clone, PartialEq, Hash)]
 pub enum Transformation {
@@ -426,7 +463,7 @@ cast! {
 #[derive(Default, Clone, Copy, Hash)]
 pub struct StyleChain<'a> {
     /// The first link of this chain.
-    head: &'a [Prehashed<Style>],
+    head: &'a [LazyHash<Style>],
     /// The remaining links in the chain.
     tail: Option<&'a Self>,
 }
@@ -437,67 +474,59 @@ impl<'a> StyleChain<'a> {
         Self { head: &root.0, tail: None }
     }
 
-    /// Make the given style list the first link of this chain.
+    /// Make the given chainable the first link of this chain.
     ///
     /// The resulting style chain contains styles from `local` as well as
     /// `self`. The ones from `local` take precedence over the ones from
     /// `self`. For folded properties `local` contributes the inner value.
-    pub fn chain<'b>(&'b self, local: &'b Styles) -> StyleChain<'b> {
-        if local.is_empty() {
-            *self
-        } else {
-            StyleChain { head: &local.0, tail: Some(self) }
-        }
+    pub fn chain<'b, C>(&'b self, local: &'b C) -> StyleChain<'b>
+    where
+        C: Chainable,
+    {
+        Chainable::chain(local, self)
+    }
+
+    /// Cast the first value for the given property in the chain.
+    pub fn get<T: Clone + 'static>(
+        self,
+        func: Element,
+        id: u8,
+        inherent: Option<&T>,
+        default: impl Fn() -> T,
+    ) -> T {
+        self.properties::<T>(func, id, inherent)
+            .next()
+            .cloned()
+            .unwrap_or_else(default)
     }
 
     /// Cast the first value for the given property in the chain,
-    /// returning a borrowed value if possible.
-    pub fn get_borrowed<T: Clone>(
+    /// returning a borrowed value.
+    pub fn get_ref<T: 'static>(
         self,
         func: Element,
         id: u8,
         inherent: Option<&'a T>,
-        default: &'static Lazy<T>,
+        default: impl Fn() -> &'a T,
     ) -> &'a T {
         self.properties::<T>(func, id, inherent)
             .next()
-            .unwrap_or_else(|| default)
+            .unwrap_or_else(default)
     }
 
-    /// Cast the first value for the given property in the chain.
-    pub fn get<T: Clone>(
+    /// Cast the first value for the given property in the chain, taking
+    /// `Fold` implementations into account.
+    pub fn get_folded<T: Fold + Clone + 'static>(
         self,
         func: Element,
         id: u8,
         inherent: Option<&T>,
-        default: &'static Lazy<T>,
+        default: impl Fn() -> T,
     ) -> T {
-        self.get_borrowed(func, id, inherent, default).clone()
-    }
-
-    /// Cast the first value for the given property in the chain.
-    pub fn get_resolve<T: Clone + Resolve>(
-        self,
-        func: Element,
-        id: u8,
-        inherent: Option<&T>,
-        default: &'static Lazy<T>,
-    ) -> T::Output {
-        self.get(func, id, inherent, default).resolve(self)
-    }
-
-    /// Cast the first value for the given property in the chain.
-    pub fn get_fold<T: Clone + Fold + 'static>(
-        self,
-        func: Element,
-        id: u8,
-        inherent: Option<&T>,
-        default: impl Fn() -> T::Output,
-    ) -> T::Output {
         fn next<T: Fold>(
             mut values: impl Iterator<Item = T>,
-            default: &impl Fn() -> T::Output,
-        ) -> T::Output {
+            default: &impl Fn() -> T,
+        ) -> T {
             values
                 .next()
                 .map(|value| value.fold(next(values, default)))
@@ -506,43 +535,8 @@ impl<'a> StyleChain<'a> {
         next(self.properties::<T>(func, id, inherent).cloned(), &default)
     }
 
-    /// Cast the first value for the given property in the chain.
-    pub fn get_resolve_fold<T>(
-        self,
-        func: Element,
-        id: u8,
-        inherent: Option<&T>,
-        default: impl Fn() -> <T::Output as Fold>::Output,
-    ) -> <T::Output as Fold>::Output
-    where
-        T: Resolve + Clone + 'static,
-        T::Output: Fold,
-    {
-        fn next<T>(
-            mut values: impl Iterator<Item = T>,
-            styles: StyleChain,
-            default: &impl Fn() -> <T::Output as Fold>::Output,
-        ) -> <T::Output as Fold>::Output
-        where
-            T: Resolve + 'static,
-            T::Output: Fold,
-        {
-            values
-                .next()
-                .map(|value| value.resolve(styles).fold(next(values, styles, default)))
-                .unwrap_or_else(default)
-        }
-
-        next(self.properties::<T>(func, id, inherent).cloned(), self, &default)
-    }
-
-    /// Iterate over all style recipes in the chain.
-    pub fn recipes(self) -> impl Iterator<Item = &'a Recipe> {
-        self.entries().filter_map(Style::recipe)
-    }
-
     /// Iterate over all values for the given property in the chain.
-    pub fn properties<T: 'static>(
+    fn properties<T: 'static>(
         self,
         func: Element,
         id: u8,
@@ -576,18 +570,18 @@ impl<'a> StyleChain<'a> {
     }
 
     /// Iterate over the entries of the chain.
-    fn entries(self) -> Entries<'a> {
+    pub fn entries(self) -> Entries<'a> {
         Entries { inner: [].as_slice().iter(), links: self.links() }
     }
 
     /// Iterate over the links of the chain.
-    fn links(self) -> Links<'a> {
+    pub fn links(self) -> Links<'a> {
         Links(Some(self))
     }
 
     /// Build owned styles from the suffix (all links beyond the `len`) of the
     /// chain.
-    fn suffix(self, len: usize) -> Styles {
+    pub fn suffix(self, len: usize) -> Styles {
         let mut suffix = Styles::new();
         let take = self.links().count().saturating_sub(len);
         for link in self.links().take(take) {
@@ -597,7 +591,7 @@ impl<'a> StyleChain<'a> {
     }
 
     /// Remove the last link from the chain.
-    fn pop(&mut self) {
+    pub fn pop(&mut self) {
         *self = self.tail.copied().unwrap_or_default();
     }
 }
@@ -622,9 +616,46 @@ impl PartialEq for StyleChain<'_> {
     }
 }
 
+/// Things that can be attached to a style chain.
+pub trait Chainable {
+    /// Attach `self` as the first link of the chain.
+    fn chain<'a>(&'a self, outer: &'a StyleChain<'_>) -> StyleChain<'a>;
+}
+
+impl Chainable for LazyHash<Style> {
+    fn chain<'a>(&'a self, outer: &'a StyleChain<'_>) -> StyleChain<'a> {
+        StyleChain {
+            head: std::slice::from_ref(self),
+            tail: Some(outer),
+        }
+    }
+}
+
+impl Chainable for [LazyHash<Style>] {
+    fn chain<'a>(&'a self, outer: &'a StyleChain<'_>) -> StyleChain<'a> {
+        if self.is_empty() {
+            *outer
+        } else {
+            StyleChain { head: self, tail: Some(outer) }
+        }
+    }
+}
+
+impl<const N: usize> Chainable for [LazyHash<Style>; N] {
+    fn chain<'a>(&'a self, outer: &'a StyleChain<'_>) -> StyleChain<'a> {
+        Chainable::chain(self.as_slice(), outer)
+    }
+}
+
+impl Chainable for Styles {
+    fn chain<'a>(&'a self, outer: &'a StyleChain<'_>) -> StyleChain<'a> {
+        Chainable::chain(self.0.as_slice(), outer)
+    }
+}
+
 /// An iterator over the entries in a style chain.
-struct Entries<'a> {
-    inner: std::slice::Iter<'a, Prehashed<Style>>,
+pub struct Entries<'a> {
+    inner: std::slice::Iter<'a, LazyHash<Style>>,
     links: Links<'a>,
 }
 
@@ -646,10 +677,10 @@ impl<'a> Iterator for Entries<'a> {
 }
 
 /// An iterator over the links of a style chain.
-struct Links<'a>(Option<StyleChain<'a>>);
+pub struct Links<'a>(Option<StyleChain<'a>>);
 
 impl<'a> Iterator for Links<'a> {
-    type Item = &'a [Prehashed<Style>];
+    type Item = &'a [LazyHash<Style>];
 
     fn next(&mut self) -> Option<Self::Item> {
         let StyleChain { head, tail } = self.0?;
@@ -658,201 +689,6 @@ impl<'a> Iterator for Links<'a> {
     }
 }
 
-/// A sequence of items with associated styles.
-#[derive(Clone, Hash)]
-pub struct StyleVec<T> {
-    items: Vec<T>,
-    styles: Vec<(Styles, usize)>,
-}
-
-impl<T> StyleVec<T> {
-    /// Whether there are any items in the sequence.
-    pub fn is_empty(&self) -> bool {
-        self.items.is_empty()
-    }
-
-    /// Number of items in the sequence.
-    pub fn len(&self) -> usize {
-        self.items.len()
-    }
-
-    /// Insert an item in the front. The item will share the style of the
-    /// current first item.
-    ///
-    /// This method has no effect if the vector is empty.
-    pub fn push_front(&mut self, item: T) {
-        if !self.styles.is_empty() {
-            self.items.insert(0, item);
-            self.styles[0].1 += 1;
-        }
-    }
-
-    /// Map the contained items.
-    pub fn map<F, U>(&self, f: F) -> StyleVec<U>
-    where
-        F: FnMut(&T) -> U,
-    {
-        StyleVec {
-            items: self.items.iter().map(f).collect(),
-            styles: self.styles.clone(),
-        }
-    }
-
-    /// Iterate over references to the contained items and associated styles.
-    pub fn iter(&self) -> impl Iterator<Item = (&T, &Styles)> + '_ {
-        self.items().zip(
-            self.styles
-                .iter()
-                .flat_map(|(map, count)| iter::repeat(map).take(*count)),
-        )
-    }
-
-    /// Iterate over the contained items.
-    pub fn items(&self) -> std::slice::Iter<'_, T> {
-        self.items.iter()
-    }
-
-    /// Extract the contained items.
-    pub fn into_items(self) -> Vec<T> {
-        self.items
-    }
-
-    /// Iterate over the contained style lists. Note that zipping this with
-    /// `items()` does not yield the same result as calling `iter()` because
-    /// this method only returns lists once that are shared by consecutive
-    /// items. This method is designed for use cases where you want to check,
-    /// for example, whether any of the lists fulfills a specific property.
-    pub fn styles(&self) -> impl Iterator<Item = &Styles> {
-        self.styles.iter().map(|(map, _)| map)
-    }
-}
-
-impl<'a> StyleVec<Cow<'a, Content>> {
-    pub fn to_vec<F: From<Content>>(self) -> Vec<F> {
-        self.items
-            .into_iter()
-            .zip(
-                self.styles
-                    .iter()
-                    .flat_map(|(map, count)| iter::repeat(map).take(*count)),
-            )
-            .map(|(content, styles)| content.into_owned().styled_with_map(styles.clone()))
-            .map(F::from)
-            .collect()
-    }
-}
-
-impl<T> Default for StyleVec<T> {
-    fn default() -> Self {
-        Self { items: vec![], styles: vec![] }
-    }
-}
-
-impl<T> FromIterator<T> for StyleVec<T> {
-    fn from_iter<I: IntoIterator<Item = T>>(iter: I) -> Self {
-        let items: Vec<_> = iter.into_iter().collect();
-        let styles = vec![(Styles::new(), items.len())];
-        Self { items, styles }
-    }
-}
-
-impl<T: Debug> Debug for StyleVec<T> {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        f.debug_list()
-            .entries(self.iter().map(|(item, styles)| {
-                crate::util::debug(|f| {
-                    styles.fmt(f)?;
-                    item.fmt(f)
-                })
-            }))
-            .finish()
-    }
-}
-
-/// Assists in the construction of a [`StyleVec`].
-#[derive(Debug)]
-pub struct StyleVecBuilder<'a, T> {
-    items: Vec<T>,
-    chains: Vec<(StyleChain<'a>, usize)>,
-}
-
-impl<'a, T> StyleVecBuilder<'a, T> {
-    /// Create a new style-vec builder.
-    pub fn new() -> Self {
-        Self { items: vec![], chains: vec![] }
-    }
-
-    /// Whether the builder is empty.
-    pub fn is_empty(&self) -> bool {
-        self.items.is_empty()
-    }
-
-    /// Push a new item into the style vector.
-    pub fn push(&mut self, item: T, styles: StyleChain<'a>) {
-        self.items.push(item);
-
-        if let Some((prev, count)) = self.chains.last_mut() {
-            if *prev == styles {
-                *count += 1;
-                return;
-            }
-        }
-
-        self.chains.push((styles, 1));
-    }
-
-    /// Iterate over the contained items.
-    pub fn elems(&self) -> std::slice::Iter<'_, T> {
-        self.items.iter()
-    }
-
-    /// Finish building, returning a pair of two things:
-    /// - a style vector of items with the non-shared styles
-    /// - a shared prefix chain of styles that apply to all items
-    pub fn finish(self) -> (StyleVec<T>, StyleChain<'a>) {
-        let mut iter = self.chains.iter();
-        let mut trunk = match iter.next() {
-            Some(&(chain, _)) => chain,
-            None => return Default::default(),
-        };
-
-        let mut shared = trunk.links().count();
-        for &(mut chain, _) in iter {
-            let len = chain.links().count();
-            if len < shared {
-                for _ in 0..shared - len {
-                    trunk.pop();
-                }
-                shared = len;
-            } else if len > shared {
-                for _ in 0..len - shared {
-                    chain.pop();
-                }
-            }
-
-            while shared > 0 && chain != trunk {
-                trunk.pop();
-                chain.pop();
-                shared -= 1;
-            }
-        }
-
-        let styles = self
-            .chains
-            .into_iter()
-            .map(|(chain, count)| (chain.suffix(shared), count))
-            .collect();
-
-        (StyleVec { items: self.items, styles }, trunk)
-    }
-}
-
-impl<'a, T> Default for StyleVecBuilder<'a, T> {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
 /// A property that is resolved with other properties from the style chain.
 pub trait Resolve {
     /// The type of the resolved output.
@@ -881,39 +717,75 @@ impl<T: Resolve> Resolve for Option<T> {
 /// #rect()
 /// ```
 pub trait Fold {
-    /// The type of the folded output.
-    type Output;
-
     /// Fold this inner value with an outer folded value.
-    fn fold(self, outer: Self::Output) -> Self::Output;
+    fn fold(self, outer: Self) -> Self;
 }
 
-impl<T> Fold for Option<T>
-where
-    T: Fold,
-    T::Output: Default,
-{
-    type Output = Option<T::Output>;
+impl Fold for bool {
+    fn fold(self, _: Self) -> Self {
+        self
+    }
+}
 
-    fn fold(self, outer: Self::Output) -> Self::Output {
-        self.map(|inner| inner.fold(outer.unwrap_or_default()))
+impl<T: Fold> Fold for Option<T> {
+    fn fold(self, outer: Self) -> Self {
+        match (self, outer) {
+            (Some(inner), Some(outer)) => Some(inner.fold(outer)),
+            // An explicit `None` should be respected, thus we don't do
+            // `inner.or(outer)`.
+            (inner, _) => inner,
+        }
     }
 }
 
 impl<T> Fold for Vec<T> {
-    type Output = Vec<T>;
-
-    fn fold(mut self, outer: Self::Output) -> Self::Output {
-        self.extend(outer);
-        self
+    fn fold(self, mut outer: Self) -> Self {
+        outer.extend(self);
+        outer
     }
 }
 
 impl<T, const N: usize> Fold for SmallVec<[T; N]> {
-    type Output = SmallVec<[T; N]>;
+    fn fold(self, mut outer: Self) -> Self {
+        outer.extend(self);
+        outer
+    }
+}
 
-    fn fold(mut self, outer: Self::Output) -> Self::Output {
-        self.extend(outer);
-        self
+/// A variant of fold for foldable optional (`Option<T>`) values where an inner
+/// `None` value isn't respected (contrary to `Option`'s usual `Fold`
+/// implementation, with which folding with an inner `None` always returns
+/// `None`). Instead, when either of the `Option` objects is `None`, the other
+/// one is necessarily returned by `fold_or`. Normal folding still occurs when
+/// both values are `Some`, using `T`'s `Fold` implementation.
+///
+/// This is useful when `None` in a particular context means "unspecified"
+/// rather than "absent", in which case a specified value (`Some`) is chosen
+/// over an unspecified one (`None`), while two specified values are folded
+/// together.
+pub trait AlternativeFold {
+    /// Attempts to fold this inner value with an outer value. However, if
+    /// either value is `None`, returns the other one instead of folding.
+    fn fold_or(self, outer: Self) -> Self;
+}
+
+impl<T: Fold> AlternativeFold for Option<T> {
+    fn fold_or(self, outer: Self) -> Self {
+        match (self, outer) {
+            (Some(inner), Some(outer)) => Some(inner.fold(outer)),
+            // If one of values is `None`, return the other one instead of
+            // folding.
+            (inner, outer) => inner.or(outer),
+        }
+    }
+}
+
+/// A type that accumulates depth when folded.
+#[derive(Debug, Default, Clone, Copy, PartialEq, Hash)]
+pub struct Depth(pub usize);
+
+impl Fold for Depth {
+    fn fold(self, outer: Self) -> Self {
+        Self(outer.0 + self.0)
     }
 }
