@@ -1,15 +1,16 @@
 use crate::diag::SourceResult;
 use crate::engine::Engine;
-use crate::foundations::{elem, Behave, Behaviour, Content, Show, StyleChain, Value};
+use crate::foundations::{elem, Content, Packed, Show, StyleChain, Value};
 use crate::introspection::Locatable;
+use crate::realize::{Behave, Behaviour};
 
 /// Exposes a value to the query system without producing visible content.
 ///
-/// This element can be retrieved with the [`query`]($query) function and from
-/// the command with [`typst query`]($reference/meta/query/#cli-queries). Its
+/// This element can be retrieved with the [`query`] function and from the
+/// command line with [`typst query`]($reference/meta/query/#cli-queries). Its
 /// purpose is to expose an arbitrary value to the introspection system. To
-/// identify a metadata value among others, you can attach a [`label`]($label)
-/// to it and query for that label.
+/// identify a metadata value among others, you can attach a [`label`] to it and
+/// query for that label.
 ///
 /// The `metadata` element is especially useful for command line queries because
 /// it allows you to expose arbitrary values to the outside world.
@@ -19,9 +20,9 @@ use crate::introspection::Locatable;
 /// #metadata("This is a note") <note>
 ///
 /// // And find it from anywhere else.
-/// #locate(loc => {
-///   query(<note>, loc).first().value
-/// })
+/// #context {
+///   query(<note>).first().value
+/// }
 /// ```
 #[elem(Behave, Show, Locatable)]
 pub struct MetadataElem {
@@ -30,13 +31,15 @@ pub struct MetadataElem {
     pub value: Value,
 }
 
-impl Show for MetadataElem {
+impl Show for Packed<MetadataElem> {
     fn show(&self, _: &mut Engine, _styles: StyleChain) -> SourceResult<Content> {
         Ok(Content::empty())
     }
 }
 
-impl Behave for MetadataElem {
+// TODO: Metadata shows as empty content, so this behaviour is unnecessary. It
+// can be removed once 0.11 has landed.
+impl Behave for Packed<MetadataElem> {
     fn behaviour(&self) -> Behaviour {
         Behaviour::Invisible
     }
