@@ -1,34 +1,32 @@
-use comemo::Prehashed;
+use std::fmt::{self, Debug, Formatter};
 
 use crate::diag::{bail, SourceResult};
 use crate::engine::Engine;
-use crate::foundations::{elem, Content, NativeElement, Resolve, Smart, StyleChain};
+use crate::foundations::{
+    elem, Content, NativeElement, Packed, Resolve, Smart, StyleChain, StyledElem,
+};
 use crate::introspection::{Meta, MetaElem};
 use crate::layout::{
-    Abs, AlignElem, Axes, BlockElem, ColbreakElem, ColumnsElem, FixedAlign, Fr, Fragment,
-    Frame, FrameItem, Layout, PlaceElem, Point, Regions, Rel, Size, Spacing, VAlign,
-    VElem,
+    Abs, AlignElem, Axes, BlockElem, ColbreakElem, ColumnsElem, FixedAlignment, Fr,
+    Fragment, Frame, FrameItem, LayoutMultiple, LayoutSingle, PlaceElem, Point, Regions,
+    Rel, Size, Spacing, VElem,
 };
 use crate::model::{FootnoteElem, FootnoteEntry, ParElem};
 use crate::util::Numeric;
-use crate::visualize::{
-    CircleElem, EllipseElem, ImageElem, LineElem, PathElem, PolygonElem, RectElem,
-    SquareElem,
-};
 
 /// Arranges spacing, paragraphs and block-level elements into a flow.
 ///
 /// This element is responsible for layouting both the top-level content flow
 /// and the contents of boxes.
-#[elem(Layout)]
+#[elem(Debug, LayoutMultiple)]
 pub struct FlowElem {
     /// The children that will be arranges into a flow.
     #[variadic]
-    pub children: Vec<Prehashed<Content>>,
+    pub children: Vec<Content>,
 }
 
-impl Layout for FlowElem {
-    #[tracing::instrument(name = "FlowElem::layout", skip_all)]
+impl LayoutMultiple for Packed<FlowElem> {
+    #[typst_macros::time(name = "flow", span = self.span())]
     fn layout(
         &self,
         engine: &mut Engine,
@@ -41,49 +39,33 @@ impl Layout for FlowElem {
         if !regions.size.y.is_finite() && regions.expand.y {
             bail!(self.span(), "cannot expand into infinite height");
         }
-        let mut layouter = FlowLayouter::new(regions, styles);
 
-        for mut child in self.children().iter().map(|c| &**c) {
+        let mut layouter = FlowLayouter::new(regions, styles);
+        for mut child in self.children().iter() {
             let outer = styles;
             let mut styles = styles;
-            if let Some((elem, map)) = child.to_styled() {
-                child = elem;
-                styles = outer.chain(map);
+            if let Some(styled) = child.to_packed::<StyledElem>() {
+                child = &styled.child;
+                styles = outer.chain(&styled.styles);
             }
 
-            if let Some(elem) = child.to::<VElem>() {
+            if child.is::<MetaElem>() {
+                layouter.layout_meta(styles);
+            } else if let Some(elem) = child.to_packed::<VElem>() {
                 layouter.layout_spacing(engine, elem, styles)?;
-            } else if let Some(elem) = child.to::<ParElem>() {
-                layouter.layout_par(engine, elem, styles)?;
-            } else if child.is::<LineElem>()
-                || child.is::<RectElem>()
-                || child.is::<SquareElem>()
-                || child.is::<EllipseElem>()
-                || child.is::<CircleElem>()
-                || child.is::<ImageElem>()
-                || child.is::<PolygonElem>()
-                || child.is::<PathElem>()
-            {
-                let layoutable = child.with::<dyn Layout>().unwrap();
-                layouter.layout_single(engine, layoutable, styles)?;
-            } else if child.is::<MetaElem>() {
-                let mut frame = Frame::soft(Size::zero());
-                frame.meta(styles, true);
-                layouter.items.push(FlowItem::Frame {
-                    frame,
-                    align: Axes::splat(FixedAlign::Start),
-                    sticky: true,
-                    movable: false,
-                });
-            } else if let Some(placed) = child.to::<PlaceElem>() {
+            } else if let Some(placed) = child.to_packed::<PlaceElem>() {
                 layouter.layout_placed(engine, placed, styles)?;
-            } else if child.can::<dyn Layout>() {
-                layouter.layout_multiple(engine, child, styles)?;
             } else if child.is::<ColbreakElem>() {
                 if !layouter.regions.backlog.is_empty() || layouter.regions.last.is_some()
                 {
-                    layouter.finish_region(engine)?;
+                    layouter.finish_region(engine, true)?;
                 }
+            } else if let Some(elem) = child.to_packed::<ParElem>() {
+                layouter.layout_par(engine, elem, styles)?;
+            } else if let Some(layoutable) = child.with::<dyn LayoutSingle>() {
+                layouter.layout_single(engine, layoutable, styles)?;
+            } else if child.can::<dyn LayoutMultiple>() {
+                layouter.layout_multiple(engine, child, styles)?;
             } else {
                 bail!(child.span(), "unexpected flow child");
             }
@@ -93,6 +75,13 @@ impl Layout for FlowElem {
     }
 }
 
+impl Debug for FlowElem {
+    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
+        write!(f, "Flow ")?;
+        f.debug_list().entries(&self.children).finish()
+    }
+}
+
 /// Performs flow layout.
 struct FlowLayouter<'a> {
     /// Whether this is the root flow.
@@ -137,12 +126,12 @@ enum FlowItem {
     /// A frame for a layouted block, how to align it, whether it sticks to the
     /// item after it (for orphan prevention), and whether it is movable
     /// (to keep it together with its footnotes).
-    Frame { frame: Frame, align: Axes<FixedAlign>, sticky: bool, movable: bool },
+    Frame { frame: Frame, align: Axes<FixedAlignment>, sticky: bool, movable: bool },
     /// An absolutely placed frame.
     Placed {
         frame: Frame,
-        x_align: FixedAlign,
-        y_align: Smart<Option<FixedAlign>>,
+        x_align: FixedAlignment,
+        y_align: Smart<Option<FixedAlignment>>,
         delta: Axes<Rel<Abs>>,
         float: bool,
         clearance: Abs,
@@ -160,6 +149,20 @@ impl FlowItem {
             Self::Frame { frame, .. } | Self::Footnote(frame) => frame.height(),
         }
     }
+
+    /// Whether this item is out-of-flow.
+    ///
+    /// Out-of-flow items are guaranteed to have a [`Size::zero()`].
+    fn is_out_of_flow(&self) -> bool {
+        match self {
+            Self::Placed { float: false, .. } => true,
+            Self::Frame { frame, .. } => {
+                frame.size().is_zero()
+                    && frame.items().all(|(_, item)| matches!(item, FrameItem::Meta(..)))
+            }
+            _ => false,
+        }
+    }
 }
 
 impl<'a> FlowLayouter<'a> {
@@ -190,12 +193,23 @@ impl<'a> FlowLayouter<'a> {
         }
     }
 
+    /// Place explicit metadata into the flow.
+    fn layout_meta(&mut self, styles: StyleChain) {
+        let mut frame = Frame::soft(Size::zero());
+        frame.meta(styles, true);
+        self.items.push(FlowItem::Frame {
+            frame,
+            align: Axes::splat(FixedAlignment::Start),
+            sticky: true,
+            movable: false,
+        });
+    }
+
     /// Layout vertical spacing.
-    #[tracing::instrument(name = "FlowLayouter::layout_spacing", skip_all)]
     fn layout_spacing(
         &mut self,
         engine: &mut Engine,
-        v: &VElem,
+        v: &Packed<VElem>,
         styles: StyleChain,
     ) -> SourceResult<()> {
         self.layout_item(
@@ -211,11 +225,10 @@ impl<'a> FlowLayouter<'a> {
     }
 
     /// Layout a paragraph.
-    #[tracing::instrument(name = "FlowLayouter::layout_par", skip_all)]
     fn layout_par(
         &mut self,
         engine: &mut Engine,
-        par: &ParElem,
+        par: &Packed<ParElem>,
         styles: StyleChain,
     ) -> SourceResult<()> {
         let align = AlignElem::alignment_in(styles).resolve(styles);
@@ -231,22 +244,28 @@ impl<'a> FlowLayouter<'a> {
             )?
             .into_frames();
 
-        let mut sticky = self.items.len();
-        for (i, item) in self.items.iter().enumerate().rev() {
-            match *item {
-                FlowItem::Absolute(_, _) => {}
-                FlowItem::Frame { sticky: true, .. } => sticky = i,
-                _ => break,
-            }
-        }
-
         if let Some(first) = lines.first() {
-            if !self.regions.size.y.fits(first.height()) && !self.regions.in_last() {
+            while !self.regions.size.y.fits(first.height()) && !self.regions.in_last() {
+                let mut sticky = self.items.len();
+                for (i, item) in self.items.iter().enumerate().rev() {
+                    match *item {
+                        FlowItem::Absolute(_, _) => {}
+                        FlowItem::Frame { sticky: true, .. } => sticky = i,
+                        _ => break,
+                    }
+                }
+
                 let carry: Vec<_> = self.items.drain(sticky..).collect();
-                self.finish_region(engine)?;
+                self.finish_region(engine, false)?;
+                let in_last = self.regions.in_last();
+
                 for item in carry {
                     self.layout_item(engine, item)?;
                 }
+
+                if in_last {
+                    break;
+                }
             }
         }
 
@@ -266,17 +285,17 @@ impl<'a> FlowLayouter<'a> {
     }
 
     /// Layout into a single region.
-    #[tracing::instrument(name = "FlowLayouter::layout_single", skip_all)]
     fn layout_single(
         &mut self,
         engine: &mut Engine,
-        content: &dyn Layout,
+        layoutable: &dyn LayoutSingle,
         styles: StyleChain,
     ) -> SourceResult<()> {
         let align = AlignElem::alignment_in(styles).resolve(styles);
         let sticky = BlockElem::sticky_in(styles);
         let pod = Regions::one(self.regions.base(), Axes::splat(false));
-        let frame = content.layout(engine, styles, pod)?.into_frame();
+        let mut frame = layoutable.layout(engine, styles, pod)?;
+        frame.meta(styles, false);
         self.layout_item(
             engine,
             FlowItem::Frame { frame, align, sticky, movable: true },
@@ -289,18 +308,19 @@ impl<'a> FlowLayouter<'a> {
     fn layout_placed(
         &mut self,
         engine: &mut Engine,
-        placed: &PlaceElem,
+        placed: &Packed<PlaceElem>,
         styles: StyleChain,
     ) -> SourceResult<()> {
         let float = placed.float(styles);
         let clearance = placed.clearance(styles);
         let alignment = placed.alignment(styles);
         let delta = Axes::new(placed.dx(styles), placed.dy(styles)).resolve(styles);
-        let x_align = alignment.map_or(FixedAlign::Center, |align| {
+        let x_align = alignment.map_or(FixedAlignment::Center, |align| {
             align.x().unwrap_or_default().resolve(styles)
         });
-        let y_align = alignment.map(|align| align.y().map(VAlign::fix));
-        let frame = placed.layout(engine, styles, self.regions)?.into_frame();
+        let y_align = alignment.map(|align| align.y().map(|y| y.resolve(styles)));
+        let mut frame = placed.layout(engine, styles, self.regions.base())?.into_frame();
+        frame.meta(styles, false);
         let item = FlowItem::Placed { frame, x_align, y_align, delta, float, clearance };
         self.layout_item(engine, item)
     }
@@ -309,12 +329,12 @@ impl<'a> FlowLayouter<'a> {
     fn layout_multiple(
         &mut self,
         engine: &mut Engine,
-        block: &Content,
+        child: &Content,
         styles: StyleChain,
     ) -> SourceResult<()> {
         // Temporarily delegerate rootness to the columns.
         let is_root = self.root;
-        if is_root && block.is::<ColumnsElem>() {
+        if is_root && child.is::<ColumnsElem>() {
             self.root = false;
             self.regions.root = true;
         }
@@ -323,14 +343,14 @@ impl<'a> FlowLayouter<'a> {
 
         if self.regions.is_full() {
             // Skip directly if region is already full.
-            self.finish_region(engine)?;
+            self.finish_region(engine, false)?;
         }
 
         // How to align the block.
-        let align = if let Some(align) = block.to::<AlignElem>() {
+        let align = if let Some(align) = child.to_packed::<AlignElem>() {
             align.alignment(styles)
-        } else if let Some((_, local)) = block.to_styled() {
-            AlignElem::alignment_in(styles.chain(local))
+        } else if let Some(styled) = child.to_packed::<StyledElem>() {
+            AlignElem::alignment_in(styles.chain(&styled.styles))
         } else {
             AlignElem::alignment_in(styles)
         }
@@ -338,20 +358,23 @@ impl<'a> FlowLayouter<'a> {
 
         // Layout the block itself.
         let sticky = BlockElem::sticky_in(styles);
-        let fragment = block.layout(engine, styles, self.regions)?;
+        let fragment = child.layout(engine, styles, self.regions)?;
 
-        for (i, frame) in fragment.into_iter().enumerate() {
+        for (i, mut frame) in fragment.into_iter().enumerate() {
             // Find footnotes in the frame.
             if self.root {
                 find_footnotes(&mut notes, &frame);
             }
 
             if i > 0 {
-                self.finish_region(engine)?;
+                self.finish_region(engine, false)?;
             }
 
-            let item = FlowItem::Frame { frame, align, sticky, movable: false };
-            self.layout_item(engine, item)?;
+            frame.meta(styles, false);
+            self.layout_item(
+                engine,
+                FlowItem::Frame { frame, align, sticky, movable: false },
+            )?;
         }
 
         self.try_handle_footnotes(engine, notes)?;
@@ -364,7 +387,6 @@ impl<'a> FlowLayouter<'a> {
     }
 
     /// Layout a finished frame.
-    #[tracing::instrument(name = "FlowLayouter::layout_item", skip_all)]
     fn layout_item(
         &mut self,
         engine: &mut Engine,
@@ -385,8 +407,8 @@ impl<'a> FlowLayouter<'a> {
             FlowItem::Fractional(_) => {}
             FlowItem::Frame { ref frame, movable, .. } => {
                 let height = frame.height();
-                if !self.regions.size.y.fits(height) && !self.regions.in_last() {
-                    self.finish_region(engine)?;
+                while !self.regions.size.y.fits(height) && !self.regions.in_last() {
+                    self.finish_region(engine, false)?;
                 }
 
                 self.regions.size.y -= height;
@@ -396,7 +418,7 @@ impl<'a> FlowLayouter<'a> {
                     self.items.push(item);
                     if !self.handle_footnotes(engine, &mut notes, true, false)? {
                         let item = self.items.pop();
-                        self.finish_region(engine)?;
+                        self.finish_region(engine, false)?;
                         self.items.extend(item);
                         self.regions.size.y -= height;
                         self.handle_footnotes(engine, &mut notes, true, true)?;
@@ -412,9 +434,11 @@ impl<'a> FlowLayouter<'a> {
                 clearance,
                 ..
             } => {
-                // If the float doesn't fit, queue it for the next region.
-                if !self.regions.size.y.fits(frame.height() + clearance)
-                    && !self.regions.in_last()
+                // If there is a queued float in front or if the float doesn't
+                // fit, queue it for the next region.
+                if !self.pending_floats.is_empty()
+                    || (!self.regions.size.y.fits(frame.height() + clearance)
+                        && !self.regions.in_last())
                 {
                     self.pending_floats.push(item);
                     return Ok(());
@@ -425,15 +449,18 @@ impl<'a> FlowLayouter<'a> {
                     let ratio = (self.regions.size.y
                         - (frame.height() + clearance) / 2.0)
                         / self.regions.full;
-                    let better_align =
-                        if ratio <= 0.5 { FixedAlign::End } else { FixedAlign::Start };
+                    let better_align = if ratio <= 0.5 {
+                        FixedAlignment::End
+                    } else {
+                        FixedAlignment::Start
+                    };
                     *y_align = Smart::Custom(Some(better_align));
                 }
 
                 // Add some clearance so that the float doesn't touch the main
                 // content.
                 frame.size_mut().y += clearance;
-                if *y_align == Smart::Custom(Some(FixedAlign::End)) {
+                if *y_align == Smart::Custom(Some(FixedAlignment::End)) {
                     frame.translate(Point::with_y(clearance));
                 }
 
@@ -454,12 +481,26 @@ impl<'a> FlowLayouter<'a> {
     }
 
     /// Finish the frame for one region.
-    fn finish_region(&mut self, engine: &mut Engine) -> SourceResult<()> {
+    ///
+    /// Set `force` to `true` to allow creating a frame for out-of-flow elements
+    /// only (this is used to force the creation of a frame in case the
+    /// remaining elements are all out-of-flow).
+    fn finish_region(&mut self, engine: &mut Engine, force: bool) -> SourceResult<()> {
+        if !force
+            && !self.items.is_empty()
+            && self.items.iter().all(FlowItem::is_out_of_flow)
+        {
+            self.finished.push(Frame::soft(self.initial));
+            self.regions.next();
+            self.initial = self.regions.size;
+            return Ok(());
+        }
+
         // Trim weak spacing.
         while self
             .items
             .last()
-            .map_or(false, |item| matches!(item, FlowItem::Absolute(_, true)))
+            .is_some_and(|item| matches!(item, FlowItem::Absolute(_, true)))
         {
             self.items.pop();
         }
@@ -481,10 +522,10 @@ impl<'a> FlowLayouter<'a> {
                 }
                 FlowItem::Placed { float: false, .. } => {}
                 FlowItem::Placed { frame, float: true, y_align, .. } => match y_align {
-                    Smart::Custom(Some(FixedAlign::Start)) => {
+                    Smart::Custom(Some(FixedAlignment::Start)) => {
                         float_top_height += frame.height()
                     }
-                    Smart::Custom(Some(FixedAlign::End)) => {
+                    Smart::Custom(Some(FixedAlignment::End)) => {
                         float_bottom_height += frame.height()
                     }
                     _ => {}
@@ -510,7 +551,7 @@ impl<'a> FlowLayouter<'a> {
         }
 
         let mut output = Frame::soft(size);
-        let mut ruler = FixedAlign::Start;
+        let mut ruler = FixedAlignment::Start;
         let mut float_top_offset = Abs::zero();
         let mut offset = float_top_height;
         let mut float_bottom_offset = Abs::zero();
@@ -538,12 +579,12 @@ impl<'a> FlowLayouter<'a> {
                     let x = x_align.position(size.x - frame.width());
                     let y = if float {
                         match y_align {
-                            Smart::Custom(Some(FixedAlign::Start)) => {
+                            Smart::Custom(Some(FixedAlignment::Start)) => {
                                 let y = float_top_offset;
                                 float_top_offset += frame.height();
                                 y
                             }
-                            Smart::Custom(Some(FixedAlign::End)) => {
+                            Smart::Custom(Some(FixedAlignment::End)) => {
                                 let y = size.y - footnote_height - float_bottom_height
                                     + float_bottom_offset;
                                 float_bottom_offset += frame.height();
@@ -579,7 +620,7 @@ impl<'a> FlowLayouter<'a> {
         self.initial = self.regions.size;
         self.has_footnotes = false;
 
-        // Try to place floats.
+        // Try to place floats into the next region.
         for item in std::mem::take(&mut self.pending_floats) {
             self.layout_item(engine, item)?;
         }
@@ -591,13 +632,13 @@ impl<'a> FlowLayouter<'a> {
     fn finish(mut self, engine: &mut Engine) -> SourceResult<Fragment> {
         if self.expand.y {
             while !self.regions.backlog.is_empty() {
-                self.finish_region(engine)?;
+                self.finish_region(engine, true)?;
             }
         }
 
-        self.finish_region(engine)?;
+        self.finish_region(engine, true)?;
         while !self.items.is_empty() {
-            self.finish_region(engine)?;
+            self.finish_region(engine, true)?;
         }
 
         Ok(Fragment::frames(self.finished))
@@ -608,21 +649,20 @@ impl FlowLayouter<'_> {
     fn try_handle_footnotes(
         &mut self,
         engine: &mut Engine,
-        mut notes: Vec<FootnoteElem>,
+        mut notes: Vec<Packed<FootnoteElem>>,
     ) -> SourceResult<()> {
         if self.root && !self.handle_footnotes(engine, &mut notes, false, false)? {
-            self.finish_region(engine)?;
+            self.finish_region(engine, false)?;
             self.handle_footnotes(engine, &mut notes, false, true)?;
         }
         Ok(())
     }
 
     /// Processes all footnotes in the frame.
-    #[tracing::instrument(skip_all)]
     fn handle_footnotes(
         &mut self,
         engine: &mut Engine,
-        notes: &mut Vec<FootnoteElem>,
+        notes: &mut Vec<Packed<FootnoteElem>>,
         movable: bool,
         force: bool,
     ) -> SourceResult<bool> {
@@ -652,7 +692,7 @@ impl FlowLayouter<'_> {
             // together).
             if !force
                 && (k == 0 || movable)
-                && frames.first().map_or(false, Frame::is_empty)
+                && frames.first().is_some_and(Frame::is_empty)
             {
                 // Remove existing footnotes attempts because we need to
                 // move the item to the next page.
@@ -673,7 +713,7 @@ impl FlowLayouter<'_> {
             for (i, frame) in frames.into_iter().enumerate() {
                 find_footnotes(notes, &frame);
                 if i > 0 {
-                    self.finish_region(engine)?;
+                    self.finish_region(engine, false)?;
                     self.layout_footnote_separator(engine)?;
                     self.regions.size.y -= self.footnote_config.gap;
                 }
@@ -695,7 +735,6 @@ impl FlowLayouter<'_> {
     }
 
     /// Layout and save the footnote separator, typically a line.
-    #[tracing::instrument(skip_all)]
     fn layout_footnote_separator(&mut self, engine: &mut Engine) -> SourceResult<()> {
         let expand = Axes::new(self.regions.expand.x, false);
         let pod = Regions::one(self.regions.base(), expand);
@@ -714,15 +753,16 @@ impl FlowLayouter<'_> {
 }
 
 /// Finds all footnotes in the frame.
-#[tracing::instrument(skip_all)]
-fn find_footnotes(notes: &mut Vec<FootnoteElem>, frame: &Frame) {
+fn find_footnotes(notes: &mut Vec<Packed<FootnoteElem>>, frame: &Frame) {
     for (_, item) in frame.items() {
         match item {
             FrameItem::Group(group) => find_footnotes(notes, &group.frame),
             FrameItem::Meta(Meta::Elem(content), _)
                 if !notes.iter().any(|note| note.location() == content.location()) =>
             {
-                let Some(footnote) = content.to::<FootnoteElem>() else { continue };
+                let Some(footnote) = content.to_packed::<FootnoteElem>() else {
+                    continue;
+                };
                 notes.push(footnote.clone());
             }
             _ => {}
