@@ -4,12 +4,10 @@ use std::fmt::{self, Debug, Formatter};
 use std::num::NonZeroUsize;
 use std::sync::Arc;
 
-use ecow::{eco_format, EcoString};
-
-use crate::foundations::{cast, dict, Dict, Repr, StyleChain, Value};
+use crate::foundations::{cast, dict, Dict, StyleChain, Value};
 use crate::introspection::{Meta, MetaElem};
 use crate::layout::{
-    Abs, Axes, Corners, FixedAlign, Length, Point, Rel, Sides, Size, Transform,
+    Abs, Axes, Corners, FixedAlignment, Length, Point, Rel, Sides, Size, Transform,
 };
 use crate::syntax::Span;
 use crate::text::TextItem;
@@ -50,7 +48,7 @@ impl Frame {
 
     /// Create a new, empty soft frame.
     ///
-    /// Panics the size is not finite.
+    /// Panics if the size is not finite.
     #[track_caller]
     pub fn soft(size: Size) -> Self {
         Self::new(size, FrameKind::Soft)
@@ -257,14 +255,18 @@ impl Frame {
         }
     }
 
-    /// Resize the frame to a new size, distributing new space according to the
-    /// given alignments.
-    pub fn resize(&mut self, target: Size, align: Axes<FixedAlign>) {
-        if self.size != target {
-            let offset = align.zip_map(target - self.size, FixedAlign::position);
-            self.size = target;
-            self.translate(offset.to_point());
+    /// Adjust the frame's size, translate the original content by an offset
+    /// computed according to the given alignments, and return the amount of
+    /// offset.
+    pub fn resize(&mut self, target: Size, align: Axes<FixedAlignment>) -> Point {
+        if self.size == target {
+            return Point::zero();
         }
+        let offset =
+            align.zip_map(target - self.size, FixedAlignment::position).to_point();
+        self.size = target;
+        self.translate(offset);
+        offset
     }
 
     /// Move the baseline and contents of the frame by an offset.
@@ -289,20 +291,32 @@ impl Frame {
     /// Attach metadata from an iterator.
     pub fn meta_iter(&mut self, iter: impl IntoIterator<Item = Meta>) {
         let mut hide = false;
-        for meta in iter {
+        let size = self.size;
+        self.prepend_multiple(iter.into_iter().filter_map(|meta| {
             if matches!(meta, Meta::Hide) {
                 hide = true;
+                None
             } else {
-                self.prepend(Point::zero(), FrameItem::Meta(meta, self.size));
+                Some((Point::zero(), FrameItem::Meta(meta, size)))
             }
-        }
+        }));
         if hide {
-            Arc::make_mut(&mut self.items).retain(|(_, item)| {
-                matches!(item, FrameItem::Group(_) | FrameItem::Meta(Meta::Elem(_), _))
-            });
+            self.hide();
         }
     }
 
+    /// Hide all content in the frame, but keep metadata.
+    pub fn hide(&mut self) {
+        Arc::make_mut(&mut self.items).retain_mut(|(_, item)| match item {
+            FrameItem::Group(group) => {
+                group.frame.hide();
+                !group.frame.is_empty()
+            }
+            FrameItem::Meta(Meta::Elem(_), _) => true,
+            _ => false,
+        });
+    }
+
     /// Add a background fill.
     pub fn fill(&mut self, fill: Paint) {
         self.prepend(
@@ -365,13 +379,13 @@ impl Frame {
 /// Tools for debugging.
 impl Frame {
     /// Add a full size aqua background and a red baseline for debugging.
-    pub fn debug(mut self) -> Self {
-        self.debug_in_place();
+    pub fn mark_box(mut self) -> Self {
+        self.mark_box_in_place();
         self
     }
 
-    /// Debug in place.
-    pub fn debug_in_place(&mut self) {
+    /// Debug in place. Add a full size aqua background and a red baseline for debugging.
+    pub fn mark_box_in_place(&mut self) {
         self.insert(
             0,
             Point::zero(),
@@ -384,11 +398,8 @@ impl Frame {
             1,
             Point::with_y(self.baseline()),
             FrameItem::Shape(
-                Geometry::Line(Point::with_x(self.size.x)).stroked(FixedStroke {
-                    paint: Color::RED.into(),
-                    thickness: Abs::pt(1.0),
-                    ..FixedStroke::default()
-                }),
+                Geometry::Line(Point::with_x(self.size.x))
+                    .stroked(FixedStroke::from_pair(Color::RED, Abs::pt(1.0))),
                 Span::detached(),
             ),
         );
@@ -411,11 +422,8 @@ impl Frame {
         self.push(
             Point::with_y(y),
             FrameItem::Shape(
-                Geometry::Line(Point::with_x(self.size.x)).stroked(FixedStroke {
-                    paint: Color::GREEN.into(),
-                    thickness: Abs::pt(1.0),
-                    ..FixedStroke::default()
-                }),
+                Geometry::Line(Point::with_x(self.size.x))
+                    .stroked(FixedStroke::from_pair(Color::GREEN, Abs::pt(1.0))),
                 Span::detached(),
             ),
         );
@@ -546,42 +554,3 @@ impl From<Position> for Dict {
         }
     }
 }
-
-/// Specification for a PDF page label.
-#[derive(Debug, Clone, PartialEq, Hash, Default)]
-pub struct PdfPageLabel {
-    /// Can be any string or none. Will always be prepended to the numbering style.
-    pub prefix: Option<EcoString>,
-    /// Based on the numbering pattern.
-    ///
-    /// If `None` or numbering is a function, the field will be empty.
-    pub style: Option<PdfPageLabelStyle>,
-    /// Offset for the page label start.
-    ///
-    /// Describes where to start counting from when setting a style.
-    /// (Has to be greater or equal than 1)
-    pub offset: Option<NonZeroUsize>,
-}
-
-impl Repr for PdfPageLabel {
-    fn repr(&self) -> EcoString {
-        eco_format!("{self:?}")
-    }
-}
-
-/// A PDF page label number style.
-#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
-pub enum PdfPageLabelStyle {
-    /// Decimal arabic numerals (1, 2, 3).
-    Arabic,
-    /// Lowercase roman numerals (i, ii, iii).
-    LowerRoman,
-    /// Uppercase roman numerals (I, II, III).
-    UpperRoman,
-    /// Lowercase letters (`a` to `z` for the first 26 pages,
-    /// `aa` to `zz` and so on for the next).
-    LowerAlpha,
-    /// Uppercase letters (`A` to `Z` for the first 26 pages,
-    /// `AA` to `ZZ` and so on for the next).
-    UpperAlpha,
-}
