@@ -6,6 +6,7 @@ use serde::{Deserialize, Serialize};
 use ttf_parser::{name_id, PlatformId, Tag};
 use unicode_segmentation::UnicodeSegmentation;
 
+use super::exceptions::find_exception;
 use crate::text::{Font, FontStretch, FontStyle, FontVariant, FontWeight};
 
 /// Metadata about a collection of fonts.
@@ -93,8 +94,8 @@ impl FontBook {
         variant: FontVariant,
         text: &str,
     ) -> Option<usize> {
-        // Find the fonts that contain the text's first char ...
-        let c = text.chars().next()?;
+        // Find the fonts that contain the text's first non-space char ...
+        let c = text.chars().find(|c| !c.is_whitespace())?;
         let ids = self
             .infos
             .iter()
@@ -206,6 +207,8 @@ impl FontInfo {
 
     /// Compute metadata for a single ttf-parser face.
     pub(super) fn from_ttf(ttf: &ttf_parser::Face) -> Option<Self> {
+        let ps_name = find_name(ttf, name_id::POST_SCRIPT_NAME);
+        let exception = ps_name.as_deref().and_then(find_exception);
         // We cannot use Name ID 16 "Typographic Family", because for some
         // fonts it groups together more than just Style / Weight / Stretch
         // variants (e.g. Display variants of Noto fonts) and then some
@@ -217,50 +220,40 @@ impl FontInfo {
         // So, instead we use Name ID 1 "Family" and trim many common
         // suffixes for which know that they just describe styling (e.g.
         // "ExtraBold").
-        //
-        // Also, for Noto fonts we use Name ID 4 "Full Name" instead,
-        // because Name ID 1 "Family" sometimes contains "Display" and
-        // sometimes doesn't for the Display variants and that mixes things
-        // up.
-        let family = {
-            let mut family = find_name(ttf, name_id::FAMILY)?;
-            if family.starts_with("Noto")
-                || family.starts_with("NewCM")
-                || family.starts_with("NewComputerModern")
-            {
-                family = find_name(ttf, name_id::FULL_NAME)?;
-            }
-            typographic_family(&family).to_string()
-        };
+        let family =
+            exception.and_then(|c| c.family.map(str::to_string)).or_else(|| {
+                let family = find_name(ttf, name_id::FAMILY)?;
+                Some(typographic_family(&family).to_string())
+            })?;
 
         let variant = {
-            let mut full = find_name(ttf, name_id::FULL_NAME).unwrap_or_default();
-            full.make_ascii_lowercase();
+            let style = exception.and_then(|c| c.style).unwrap_or_else(|| {
+                let mut full = find_name(ttf, name_id::FULL_NAME).unwrap_or_default();
+                full.make_ascii_lowercase();
 
-            // Some fonts miss the relevant bits for italic or oblique, so
-            // we also try to infer that from the full name.
-            let italic = ttf.is_italic() || full.contains("italic");
-            let oblique =
-                ttf.is_oblique() || full.contains("oblique") || full.contains("slanted");
+                // Some fonts miss the relevant bits for italic or oblique, so
+                // we also try to infer that from the full name.
+                let italic = ttf.is_italic() || full.contains("italic");
+                let oblique = ttf.is_oblique()
+                    || full.contains("oblique")
+                    || full.contains("slanted");
 
-            let style = match (italic, oblique) {
-                (false, false) => FontStyle::Normal,
-                (true, _) => FontStyle::Italic,
-                (_, true) => FontStyle::Oblique,
-            };
-
-            let weight = {
-                let mut number = ttf.weight().to_number();
-                if (family.starts_with("NewCM")
-                    || family.starts_with("New Computer Modern"))
-                    && full.contains("book")
-                {
-                    number += 50;
+                match (italic, oblique) {
+                    (false, false) => FontStyle::Normal,
+                    (true, _) => FontStyle::Italic,
+                    (_, true) => FontStyle::Oblique,
                 }
+            });
+
+            let weight = exception.and_then(|c| c.weight).unwrap_or_else(|| {
+                let number = ttf.weight().to_number();
                 FontWeight::from_number(number)
-            };
+            });
+
+            let stretch = exception
+                .and_then(|c| c.stretch)
+                .unwrap_or_else(|| FontStretch::from_number(ttf.width().to_number()));
 
-            let stretch = FontStretch::from_number(ttf.width().to_number());
             FontVariant { style, weight, stretch }
         };
 
@@ -355,12 +348,6 @@ fn typographic_family(mut family: &str) -> &str {
         "narrow", "condensed", "cond", "cn", "cd", "compressed", "expanded", "exp"
     ];
 
-    let mut extra = [].as_slice();
-    let newcm = family.starts_with("NewCM") || family.starts_with("NewComputerModern");
-    if newcm {
-        extra = &["book"];
-    }
-
     // Trim spacing and weird leading dots in Apple fonts.
     family = family.trim().trim_start_matches('.');
 
@@ -376,7 +363,7 @@ fn typographic_family(mut family: &str) -> &str {
         // Find style suffix.
         let mut t = trimmed;
         let mut shortened = false;
-        while let Some(s) = SUFFIXES.iter().chain(extra).find_map(|s| t.strip_suffix(s)) {
+        while let Some(s) = SUFFIXES.iter().find_map(|s| t.strip_suffix(s)) {
             shortened = true;
             t = s;
         }
@@ -403,20 +390,7 @@ fn typographic_family(mut family: &str) -> &str {
     // Apply style suffix trimming.
     family = &family[..len];
 
-    if newcm {
-        family = family.trim_end_matches("10");
-    }
-
-    // Fix bad names.
-    match family {
-        "Noto Sans Symbols2" => "Noto Sans Symbols 2",
-        "NewComputerModern" => "New Computer Modern",
-        "NewComputerModernMono" => "New Computer Modern Mono",
-        "NewComputerModernSans" => "New Computer Modern Sans",
-        "NewComputerModernMath" => "New Computer Modern Math",
-        "NewCMUncial" | "NewComputerModernUncial" => "New Computer Modern Uncial",
-        other => other,
-    }
+    family
 }
 
 /// How many words the two strings share in their prefix.
