@@ -1,87 +1,126 @@
-#![allow(clippy::comparison_chain)]
+/*! This is Typst's test runner.
 
-use std::cell::{RefCell, RefMut};
+Tests are Typst files composed of a header part followed by subtests.
+
+The header may contain:
+- a small description `// tests that features X works well`
+- metadata (see [metadata::TestConfiguration])
+
+The subtests may use extra testing functions defined in [library], most
+importantly, `test(x, y)` which will fail the test `if x != y`.
+*/
+
+#![allow(clippy::comparison_chain)]
+mod metadata;
+
+use self::metadata::*;
+
+use std::borrow::Cow;
 use std::collections::{HashMap, HashSet};
 use std::ffi::OsStr;
-use std::fmt::{self, Display, Formatter, Write as _};
-use std::io::{self, Write};
+use std::fmt::Write as _;
+use std::io::{self, IsTerminal, Write as _};
 use std::ops::Range;
-use std::path::{Path, PathBuf};
+use std::path::{Path, PathBuf, MAIN_SEPARATOR_STR};
+use std::sync::{OnceLock, RwLock};
 use std::{env, fs};
 
 use clap::Parser;
 use comemo::{Prehashed, Track};
-use ecow::EcoString;
 use oxipng::{InFile, Options, OutFile};
 use rayon::iter::{ParallelBridge, ParallelIterator};
-use std::cell::OnceCell;
 use tiny_skia as sk;
-use typst::diag::{bail, FileError, FileResult, Severity, StrResult};
+use typst::diag::{bail, FileError, FileResult, Severity, SourceDiagnostic, StrResult};
 use typst::eval::Tracer;
-use typst::foundations::{
-    eco_format, func, Bytes, Datetime, NoneValue, Repr, Smart, Value,
-};
+use typst::foundations::{func, Bytes, Datetime, NoneValue, Repr, Smart, Value};
 use typst::introspection::Meta;
-use typst::layout::{Abs, Frame, FrameItem, Margin, PageElem, Transform};
+use typst::layout::{Abs, Frame, FrameItem, Margin, Page, PageElem, Transform};
 use typst::model::Document;
-use typst::syntax::{FileId, PackageVersion, Source, SyntaxNode, VirtualPath};
+use typst::syntax::{FileId, Source, SyntaxNode, VirtualPath};
 use typst::text::{Font, FontBook, TextElem, TextSize};
 use typst::visualize::Color;
 use typst::{Library, World, WorldExt};
-use unscanny::Scanner;
 use walkdir::WalkDir;
 
+// These directories are all relative to the tests/ directory.
 const TYP_DIR: &str = "typ";
 const REF_DIR: &str = "ref";
 const PNG_DIR: &str = "png";
 const PDF_DIR: &str = "pdf";
 const SVG_DIR: &str = "svg";
-const FONT_DIR: &str = "../assets/fonts";
-const ASSET_DIR: &str = "../assets";
 
+/// Arguments that modify test behaviour.
+///
+/// Specify them like this when developing:
+/// `cargo test --workspace --test tests -- --help`
 #[derive(Debug, Clone, Parser)]
 #[clap(name = "typst-test", author)]
 struct Args {
+    /// All the tests that contains a filter string will be run (unless
+    /// `--exact` is specified, which is even stricter).
     filter: Vec<String>,
-    /// runs only the specified subtest
+    /// Runs only the specified subtest.
     #[arg(short, long)]
     #[arg(allow_hyphen_values = true)]
     subtest: Option<isize>,
+    /// Runs only the test with the exact name specified in your command.
+    ///
+    /// Example:
+    /// `cargo test --workspace --test tests  -- compiler/bytes.typ --exact`
     #[arg(long)]
     exact: bool,
+    /// Updates the reference images in `tests/ref`.
     #[arg(long, default_value_t = env::var_os("UPDATE_EXPECT").is_some())]
     update: bool,
+    /// Exports the tests as PDF into `tests/pdf`.
     #[arg(long)]
     pdf: bool,
+    /// Configuration of what to print.
     #[command(flatten)]
     print: PrintConfig,
+    /// Running `cargo test --workspace -- --nocapture` for the unit tests would
+    /// fail the test runner without argument.
+    // TODO: would it really still happen?
     #[arg(long)]
-    nocapture: bool, // simply ignores the argument
+    nocapture: bool,
+    /// Prevents the terminal from being cleared of test names and includes
+    /// non-essential test messages.
+    #[arg(short, long)]
+    verbose: bool,
 }
 
 /// Which things to print out for debugging.
 #[derive(Default, Debug, Copy, Clone, Eq, PartialEq, Parser)]
 struct PrintConfig {
+    /// Print the syntax tree.
     #[arg(long)]
     syntax: bool,
+    /// Print the content model.
     #[arg(long)]
     model: bool,
+    /// Print the layouted frames.
     #[arg(long)]
     frames: bool,
 }
 
 impl Args {
-    fn matches(&self, path: &Path) -> bool {
-        if self.exact {
-            let name = path.file_name().unwrap().to_string_lossy();
-            self.filter.iter().any(|v| v == &name)
-        } else {
-            let path = path.to_string_lossy();
-            self.filter.is_empty() || self.filter.iter().any(|v| path.contains(v))
+    fn matches(&self, canonicalized_path: &Path) -> bool {
+        let path = canonicalized_path.to_string_lossy();
+        if !self.exact {
+            return self.filter.is_empty()
+                || self.filter.iter().any(|v| path.contains(v));
         }
+
+        self.filter.iter().any(|v| match path.strip_suffix(v) {
+            None => false,
+            Some(residual) => {
+                residual.is_empty() || residual.ends_with(MAIN_SEPARATOR_STR)
+            }
+        })
     }
 }
 
+/// Tests all test files and prints a summary.
 fn main() {
     let args = Args::parse();
 
@@ -89,7 +128,8 @@ fn main() {
     let world = TestWorld::new(args.print);
 
     println!("Running tests...");
-    let results = WalkDir::new("typ")
+    let results = WalkDir::new(TYP_DIR)
+        .sort_by_file_name()
         .into_iter()
         .par_bridge()
         .filter_map(|entry| {
@@ -102,12 +142,12 @@ fn main() {
                 return None;
             }
 
-            let src_path = entry.into_path();
+            let src_path = entry.into_path(); // Relative to TYP_DIR.
             if src_path.extension() != Some(OsStr::new("typ")) {
                 return None;
             }
 
-            if args.matches(&src_path) {
+            if args.matches(&src_path.canonicalize().unwrap()) {
                 Some(src_path)
             } else {
                 None
@@ -135,8 +175,10 @@ fn main() {
 
     let len = results.len();
     let ok = results.iter().sum::<usize>();
-    if len > 1 {
-        println!("{ok} / {len} tests passed.");
+    if len > 0 {
+        println!("{ok} / {len} test{} passed.", if len > 1 { "s" } else { "" });
+    } else {
+        println!("No test ran.");
     }
 
     if ok != len {
@@ -185,7 +227,7 @@ fn library() -> Library {
     // Set page width to 120pt with 10pt margins, so that the inner page is
     // exactly 100pt wide. Page height is unbounded and font size is 10pt so
     // that it multiplies to nice round numbers.
-    let mut lib = Library::build();
+    let mut lib = Library::default();
     lib.styles
         .set(PageElem::set_width(Smart::Custom(Abs::pt(120.0).into())));
     lib.styles.set(PageElem::set_height(Smart::Auto));
@@ -209,35 +251,27 @@ fn library() -> Library {
 }
 
 /// A world that provides access to the tests environment.
-#[derive(Clone)]
 struct TestWorld {
     print: PrintConfig,
     main: FileId,
     library: Prehashed<Library>,
     book: Prehashed<FontBook>,
     fonts: Vec<Font>,
-    paths: RefCell<HashMap<FileId, PathSlot>>,
+    slots: RwLock<HashMap<FileId, FileSlot>>,
 }
 
 #[derive(Clone)]
-struct PathSlot {
-    source: OnceCell<FileResult<Source>>,
-    buffer: OnceCell<FileResult<Bytes>>,
+struct FileSlot {
+    source: OnceLock<FileResult<Source>>,
+    buffer: OnceLock<FileResult<Bytes>>,
 }
 
 impl TestWorld {
     fn new(print: PrintConfig) -> Self {
-        // Search for fonts.
-        let mut fonts = vec![];
-        for entry in WalkDir::new(FONT_DIR)
-            .sort_by_file_name()
-            .into_iter()
-            .filter_map(|e| e.ok())
-            .filter(|entry| entry.file_type().is_file())
-        {
-            let data = fs::read(entry.path()).unwrap();
-            fonts.extend(Font::iter(data.into()));
-        }
+        let fonts: Vec<_> = typst_assets::fonts()
+            .chain(typst_dev_assets::fonts())
+            .flat_map(|data| Font::iter(Bytes::from_static(data)))
+            .collect();
 
         Self {
             print,
@@ -245,7 +279,7 @@ impl TestWorld {
             library: Prehashed::new(library()),
             book: Prehashed::new(FontBook::from_fonts(&fonts)),
             fonts,
-            paths: RefCell::default(),
+            slots: RwLock::new(HashMap::new()),
         }
     }
 }
@@ -264,21 +298,28 @@ impl World for TestWorld {
     }
 
     fn source(&self, id: FileId) -> FileResult<Source> {
-        let slot = self.slot(id)?;
-        slot.source
-            .get_or_init(|| {
-                let buf = read(&system_path(id)?)?;
-                let text = String::from_utf8(buf)?;
-                Ok(Source::new(id, text))
-            })
-            .clone()
+        self.slot(id, |slot| {
+            slot.source
+                .get_or_init(|| {
+                    let buf = read(&system_path(id)?)?;
+                    let text = String::from_utf8(buf.into_owned())?;
+                    Ok(Source::new(id, text))
+                })
+                .clone()
+        })
     }
 
     fn file(&self, id: FileId) -> FileResult<Bytes> {
-        let slot = self.slot(id)?;
-        slot.buffer
-            .get_or_init(|| read(&system_path(id)?).map(Bytes::from))
-            .clone()
+        self.slot(id, |slot| {
+            slot.buffer
+                .get_or_init(|| {
+                    read(&system_path(id)?).map(|cow| match cow {
+                        Cow::Owned(buf) => buf.into(),
+                        Cow::Borrowed(buf) => Bytes::from_static(buf),
+                    })
+                })
+                .clone()
+        })
     }
 
     fn font(&self, id: usize) -> Option<Font> {
@@ -293,22 +334,37 @@ impl World for TestWorld {
 impl TestWorld {
     fn set(&mut self, path: &Path, text: String) -> Source {
         self.main = FileId::new(None, VirtualPath::new(path));
-        let mut slot = self.slot(self.main).unwrap();
         let source = Source::new(self.main, text);
-        slot.source = OnceCell::from(Ok(source.clone()));
-        source
+        self.slot(self.main, |slot| {
+            slot.source = OnceLock::from(Ok(source.clone()));
+            source
+        })
     }
 
-    fn slot(&self, id: FileId) -> FileResult<RefMut<PathSlot>> {
-        Ok(RefMut::map(self.paths.borrow_mut(), |paths| {
-            paths.entry(id).or_insert_with(|| PathSlot {
-                source: OnceCell::new(),
-                buffer: OnceCell::new(),
-            })
+    fn slot<F, T>(&self, id: FileId, f: F) -> T
+    where
+        F: FnOnce(&mut FileSlot) -> T,
+    {
+        f(self.slots.write().unwrap().entry(id).or_insert_with(|| FileSlot {
+            source: OnceLock::new(),
+            buffer: OnceLock::new(),
         }))
     }
 }
 
+impl Clone for TestWorld {
+    fn clone(&self) -> Self {
+        Self {
+            print: self.print,
+            main: self.main,
+            library: self.library.clone(),
+            book: self.book.clone(),
+            fonts: self.fonts.clone(),
+            slots: RwLock::new(self.slots.read().unwrap().clone()),
+        }
+    }
+}
+
 /// The file system path for a file ID.
 fn system_path(id: FileId) -> FileResult<PathBuf> {
     let root: PathBuf = match id.package() {
@@ -320,22 +376,28 @@ fn system_path(id: FileId) -> FileResult<PathBuf> {
 }
 
 /// Read a file.
-fn read(path: &Path) -> FileResult<Vec<u8>> {
+fn read(path: &Path) -> FileResult<Cow<'static, [u8]>> {
     // Basically symlinks `assets/files` to `tests/files` so that the assets
     // are within the test project root.
-    let mut resolved = path.to_path_buf();
-    if path.starts_with("files/") {
-        resolved = Path::new(ASSET_DIR).join(path);
+    let resolved = path.to_path_buf();
+    if let Ok(suffix) = path.strip_prefix("assets/") {
+        return typst_dev_assets::get(&suffix.to_string_lossy())
+            .map(Cow::Borrowed)
+            .ok_or_else(|| FileError::NotFound(path.into()));
     }
 
     let f = |e| FileError::from_io(e, path);
     if fs::metadata(&resolved).map_err(f)?.is_dir() {
         Err(FileError::IsDirectory)
     } else {
-        fs::read(&resolved).map_err(f)
+        fs::read(&resolved).map(Cow::Owned).map_err(f)
     }
 }
 
+/// Tests a test file and prints the result.
+///
+/// Also tests that the header of each test is written correctly.
+/// See [parse_part_metadata] for more details.
 fn test(
     world: &mut TestWorld,
     src_path: &Path,
@@ -361,10 +423,9 @@ fn test(
     let mut output = String::new();
     let mut ok = true;
     let mut updated = false;
-    let mut frames = vec![];
+    let mut pages = vec![];
     let mut line = 0;
-    let mut compare_ref = None;
-    let mut validate_hints = None;
+    let mut header_configuration = None;
     let mut compare_ever = false;
     let mut rng = LinearShift::new();
 
@@ -391,9 +452,28 @@ fn test(
                 .all(|s| s.starts_with("//") || s.chars().all(|c| c.is_whitespace()));
 
         if is_header {
-            for line in part.lines() {
-                compare_ref = get_flag_metadata(line, "Ref").or(compare_ref);
-                validate_hints = get_flag_metadata(line, "Hints").or(validate_hints);
+            let source = Source::detached(part.to_string());
+            let metadata = parse_part_metadata(&source, true);
+            match metadata {
+                Ok(metadata) => {
+                    header_configuration = Some(metadata.config);
+                }
+                Err(invalid_data) => {
+                    ok = false;
+                    writeln!(
+                        output,
+                        " Test {}: invalid metadata in header, failing the test:",
+                        name.display()
+                    )
+                    .unwrap();
+                    InvalidMetadata::write(
+                        invalid_data,
+                        &mut output,
+                        &mut |annotation, output| {
+                            print_annotation(output, &source, line, annotation)
+                        },
+                    );
+                }
             }
         } else {
             let (part_ok, compare_here, part_frames) = test_part(
@@ -401,27 +481,27 @@ fn test(
                 world,
                 src_path,
                 part.into(),
-                i,
-                compare_ref.unwrap_or(true),
-                validate_hints.unwrap_or(true),
                 line,
+                i,
+                header_configuration.as_ref().unwrap_or(&Default::default()),
                 &mut rng,
+                args.verbose,
             );
 
             ok &= part_ok;
             compare_ever |= compare_here;
-            frames.extend(part_frames);
+            pages.extend(part_frames);
         }
 
         line += part.lines().count() + 1;
     }
 
-    let document = Document { pages: frames, ..Default::default() };
+    let document = Document { pages, ..Default::default() };
     if compare_ever {
         if let Some(pdf_path) = pdf_path {
             let pdf_data = typst_pdf::pdf(
                 &document,
-                Some(&format!("typst-test: {}", name.display())),
+                Smart::Custom(&format!("typst-test: {}", name.display())),
                 world.today(Some(0)),
             );
             fs::create_dir_all(pdf_path.parent().unwrap()).unwrap();
@@ -434,11 +514,12 @@ fn test(
             }
         }
 
-        let canvas = render(&document.pages);
+        let canvas = render(&document);
         fs::create_dir_all(png_path.parent().unwrap()).unwrap();
         canvas.save_png(png_path).unwrap();
 
-        let svg = typst_svg::svg_merged(&document.pages, Abs::pt(5.0));
+        let svg = typst_svg::svg_merged(&document, Abs::pt(5.0));
+
         fs::create_dir_all(svg_path.parent().unwrap()).unwrap();
         std::fs::write(svg_path, svg.as_bytes()).unwrap();
 
@@ -475,6 +556,12 @@ fn test(
         stdout.write_all(name.to_string_lossy().as_bytes()).unwrap();
         if ok {
             writeln!(stdout, " ✔").unwrap();
+            // Don't clear the line when in verbose mode or when the reference image
+            // was updated, to show in the output which test had its image updated.
+            if !updated && !args.verbose && stdout.is_terminal() {
+                // ANSI escape codes: cursor moves up and clears the line.
+                write!(stdout, "\x1b[1A\x1b[2K").unwrap();
+            }
         } else {
             writeln!(stdout, " ❌").unwrap();
         }
@@ -489,14 +576,6 @@ fn test(
     ok
 }
 
-fn get_metadata<'a>(line: &'a str, key: &str) -> Option<&'a str> {
-    line.strip_prefix(eco_format!("// {key}: ").as_str())
-}
-
-fn get_flag_metadata(line: &str, key: &str) -> Option<bool> {
-    get_metadata(line, key).map(|value| value == "true")
-}
-
 fn update_image(png_path: &Path, ref_path: &Path) {
     oxipng::optimize(
         &InFile::Path(png_path.to_owned()),
@@ -512,35 +591,19 @@ fn test_part(
     world: &mut TestWorld,
     src_path: &Path,
     text: String,
-    i: usize,
-    compare_ref: bool,
-    validate_hints: bool,
     line: usize,
+    i: usize,
+    header_configuration: &TestConfig,
     rng: &mut LinearShift,
-) -> (bool, bool, Vec<Frame>) {
-    let mut ok = true;
-
+    verbose: bool,
+) -> (bool, bool, Vec<Page>) {
     let source = world.set(src_path, text);
     if world.print.syntax {
         writeln!(output, "Syntax Tree:\n{:#?}\n", source.root()).unwrap();
     }
 
-    let metadata = parse_part_metadata(&source);
-    let compare_ref = metadata.part_configuration.compare_ref.unwrap_or(compare_ref);
-    let validate_hints =
-        metadata.part_configuration.validate_hints.unwrap_or(validate_hints);
-
-    ok &= test_spans(output, source.root());
-    ok &= test_reparse(output, source.text(), i, rng);
-
     if world.print.model {
-        let world = (world as &dyn World).track();
-        let route = typst::engine::Route::default();
-        let mut tracer = typst::eval::Tracer::new();
-
-        let module =
-            typst::eval::eval(world, route.track(), tracer.track_mut(), &source).unwrap();
-        writeln!(output, "Model:\n{:#?}\n", module.content()).unwrap();
+        print_model(world, &source, output);
     }
 
     let mut tracer = Tracer::new();
@@ -553,11 +616,186 @@ fn test_part(
         }
     };
 
-    // Don't retain frames if we don't want to compare with reference images.
-    if !compare_ref {
-        frames.clear();
-    }
+    let metadata = parse_part_metadata(&source, false);
+    match metadata {
+        Ok(metadata) => {
+            let mut ok = true;
+            let compare_ref = metadata
+                .config
+                .compare_ref
+                .unwrap_or(header_configuration.compare_ref.unwrap_or(true));
+            let validate_hints = metadata
+                .config
+                .validate_hints
+                .unwrap_or(header_configuration.validate_hints.unwrap_or(true));
+            let validate_autocomplete = metadata
+                .config
+                .validate_autocomplete
+                .unwrap_or(header_configuration.validate_autocomplete.unwrap_or(false));
 
+            if verbose {
+                writeln!(output, "Subtest {i} runs with compare_ref={compare_ref}; validate_hints={validate_hints}; validate_autocomplete={validate_autocomplete};").unwrap();
+            }
+            ok &= test_spans(output, source.root());
+            ok &= test_reparse(output, source.text(), i, rng);
+
+            // Don't retain frames if we don't want to compare with reference images.
+            if !compare_ref {
+                frames.clear();
+            }
+
+            // we never check autocomplete and error at the same time
+
+            let diagnostic_annotations = metadata
+                .annotations
+                .iter()
+                .filter(|a| {
+                    !matches!(
+                        a.kind,
+                        AnnotationKind::AutocompleteContains
+                            | AnnotationKind::AutocompleteExcludes
+                    )
+                })
+                .cloned()
+                .collect::<HashSet<_>>();
+
+            if validate_autocomplete {
+                // warns and ignores diagnostics
+                if !diagnostic_annotations.is_empty() {
+                    writeln!(
+                        output,
+                        "  Subtest {i} contains diagnostics but is in autocomplete mode."
+                    )
+                    .unwrap();
+                    for annotation in diagnostic_annotations {
+                        write!(output, "    Ignored | ").unwrap();
+                        print_annotation(output, &source, line, &annotation);
+                    }
+                }
+
+                test_autocomplete(
+                    output,
+                    world,
+                    &source,
+                    line,
+                    i,
+                    &mut ok,
+                    metadata.annotations.iter(),
+                );
+            } else {
+                test_diagnostics(
+                    output,
+                    world,
+                    &source,
+                    line,
+                    i,
+                    &mut ok,
+                    validate_hints,
+                    diagnostics.iter(),
+                    &diagnostic_annotations,
+                );
+            }
+
+            (ok, compare_ref, frames)
+        }
+        Err(invalid_data) => {
+            writeln!(output, "  Subtest {i} has invalid metadata, failing the test:")
+                .unwrap();
+            InvalidMetadata::write(
+                invalid_data,
+                output,
+                &mut |annotation: &Annotation, output: &mut String| {
+                    print_annotation(output, &source, line, annotation)
+                },
+            );
+
+            (false, false, frames)
+        }
+    }
+}
+
+#[allow(clippy::too_many_arguments)]
+fn test_autocomplete<'a>(
+    output: &mut String,
+    world: &mut TestWorld,
+    source: &Source,
+    line: usize,
+    i: usize,
+    ok: &mut bool,
+    annotations: impl Iterator<Item = &'a Annotation>,
+) {
+    for annotation in annotations.filter(|a| {
+        matches!(
+            a.kind,
+            AnnotationKind::AutocompleteContains | AnnotationKind::AutocompleteExcludes
+        )
+    }) {
+        // Ok cause we checked in parsing that range was Some for this annotation
+        let cursor = annotation.range.as_ref().unwrap().start;
+
+        // todo, use document if is_some to test labels autocomplete
+        let completions = typst_ide::autocomplete(world, None, source, cursor, true)
+            .map(|(_, c)| c)
+            .unwrap_or_default()
+            .into_iter()
+            .map(|c| c.label.to_string())
+            .collect::<HashSet<_>>();
+        let completions =
+            completions.iter().map(|s| s.as_str()).collect::<HashSet<&str>>();
+
+        let must_contain_or_exclude = parse_string_list(&annotation.text);
+        let missing =
+            must_contain_or_exclude.difference(&completions).collect::<Vec<_>>();
+
+        if !missing.is_empty()
+            && matches!(annotation.kind, AnnotationKind::AutocompleteContains)
+        {
+            writeln!(output, "  Subtest {i} does not match expected completions.")
+                .unwrap();
+            write!(output, "  for annotation  | ").unwrap();
+            print_annotation(output, source, line, annotation);
+
+            write!(output, "    Not contained  // ").unwrap();
+            for item in missing {
+                write!(output, "{item:?}, ").unwrap()
+            }
+            writeln!(output).unwrap();
+            *ok = false;
+        }
+
+        let undesired =
+            must_contain_or_exclude.intersection(&completions).collect::<Vec<_>>();
+
+        if !undesired.is_empty()
+            && matches!(annotation.kind, AnnotationKind::AutocompleteExcludes)
+        {
+            writeln!(output, "  Subtest {i} does not match expected completions.")
+                .unwrap();
+            write!(output, "  for annotation  | ").unwrap();
+            print_annotation(output, source, line, annotation);
+
+            write!(output, "    Not excluded  // ").unwrap();
+            for item in undesired {
+                write!(output, "{item:?}, ").unwrap()
+            }
+            writeln!(output).unwrap();
+            *ok = false;
+        }
+    }
+}
+
+#[allow(clippy::too_many_arguments)]
+fn test_diagnostics<'a>(
+    output: &mut String,
+    world: &mut TestWorld,
+    source: &Source,
+    line: usize,
+    i: usize,
+    ok: &mut bool,
+    validate_hints: bool,
+    diagnostics: impl Iterator<Item = &'a SourceDiagnostic>,
+    diagnostic_annotations: &HashSet<Annotation>,
+) {
     // Map diagnostics to range and message format, discard traces and errors from
     // other files, collect hints.
     //
@@ -565,9 +803,9 @@ fn test_part(
     // verify if a hint belongs to a diagnostic or not. That should be irrelevant
     // however, as the line of the hint is still verified.
     let mut actual_diagnostics = HashSet::new();
-    for diagnostic in &diagnostics {
+    for diagnostic in diagnostics {
         // Ignore diagnostics from other files.
-        if diagnostic.span.id().map_or(false, |id| id != source.id()) {
+        if diagnostic.span.id().is_some_and(|id| id != source.id()) {
             continue;
         }
 
@@ -577,14 +815,14 @@ fn test_part(
                 Severity::Warning => AnnotationKind::Warning,
             },
             range: world.range(diagnostic.span),
-            message: diagnostic.message.replace("\\", "/"),
+            text: diagnostic.message.replace("\\", "/"),
         };
 
         if validate_hints {
             for hint in &diagnostic.hints {
                 actual_diagnostics.insert(Annotation {
                     kind: AnnotationKind::Hint,
-                    message: hint.clone(),
+                    text: hint.clone(),
                     range: annotation.range.clone(),
                 });
             }
@@ -595,10 +833,9 @@ fn test_part(
 
     // Basically symmetric_difference, but we need to know where an item is coming from.
     let mut unexpected_outputs = actual_diagnostics
-        .difference(&metadata.annotations)
+        .difference(diagnostic_annotations)
         .collect::<Vec<_>>();
-    let mut missing_outputs = metadata
-        .annotations
+    let mut missing_outputs = diagnostic_annotations
         .difference(&actual_diagnostics)
         .collect::<Vec<_>>();
 
@@ -609,20 +846,28 @@ fn test_part(
     // Is this reasonable or subject to change?
     if !(unexpected_outputs.is_empty() && missing_outputs.is_empty()) {
         writeln!(output, "  Subtest {i} does not match expected errors.").unwrap();
-        ok = false;
+        *ok = false;
 
         for unexpected in unexpected_outputs {
-            write!(output, "    Not annotated | ").unwrap();
-            print_annotation(output, &source, line, unexpected)
+            write!(output, "    Not annotated // ").unwrap();
+            print_annotation(output, source, line, unexpected)
         }
 
         for missing in missing_outputs {
-            write!(output, "    Not emitted   | ").unwrap();
-            print_annotation(output, &source, line, missing)
+            write!(output, "    Not emitted   // ").unwrap();
+            print_annotation(output, source, line, missing)
         }
     }
+}
 
-    (ok, compare_ref, frames)
+fn print_model(world: &mut TestWorld, source: &Source, output: &mut String) {
+    let world = (world as &dyn World).track();
+    let route = typst::engine::Route::default();
+    let mut tracer = typst::eval::Tracer::new();
+
+    let module =
+        typst::eval::eval(world, route.track(), tracer.track_mut(), source).unwrap();
+    writeln!(output, "Model:\n{:#?}\n", module.content()).unwrap();
 }
 
 fn print_annotation(
@@ -631,116 +876,16 @@ fn print_annotation(
     line: usize,
     annotation: &Annotation,
 ) {
-    let Annotation { range, message, kind } = annotation;
+    let Annotation { range, text, kind } = annotation;
     write!(output, "{kind}: ").unwrap();
     if let Some(range) = range {
         let start_line = 1 + line + source.byte_to_line(range.start).unwrap();
         let start_col = 1 + source.byte_to_column(range.start).unwrap();
         let end_line = 1 + line + source.byte_to_line(range.end).unwrap();
         let end_col = 1 + source.byte_to_column(range.end).unwrap();
-        write!(output, "{start_line}:{start_col}-{end_line}:{end_col}: ").unwrap();
-    }
-    writeln!(output, "{message}").unwrap();
-}
-
-struct TestConfiguration {
-    compare_ref: Option<bool>,
-    validate_hints: Option<bool>,
-}
-
-struct TestPartMetadata {
-    part_configuration: TestConfiguration,
-    annotations: HashSet<Annotation>,
-}
-
-#[derive(Debug, Clone, Eq, PartialEq, Hash)]
-struct Annotation {
-    range: Option<Range<usize>>,
-    message: EcoString,
-    kind: AnnotationKind,
-}
-
-#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
-enum AnnotationKind {
-    Error,
-    Warning,
-    Hint,
-}
-
-impl AnnotationKind {
-    fn iter() -> impl Iterator<Item = Self> {
-        [AnnotationKind::Error, AnnotationKind::Warning, AnnotationKind::Hint].into_iter()
-    }
-
-    fn as_str(self) -> &'static str {
-        match self {
-            AnnotationKind::Error => "Error",
-            AnnotationKind::Warning => "Warning",
-            AnnotationKind::Hint => "Hint",
-        }
-    }
-}
-
-impl Display for AnnotationKind {
-    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
-        f.pad(self.as_str())
-    }
-}
-
-fn parse_part_metadata(source: &Source) -> TestPartMetadata {
-    let mut compare_ref = None;
-    let mut validate_hints = None;
-    let mut annotations = HashSet::default();
-
-    let lines: Vec<_> = source.text().lines().map(str::trim).collect();
-    for (i, line) in lines.iter().enumerate() {
-        compare_ref = get_flag_metadata(line, "Ref").or(compare_ref);
-        validate_hints = get_flag_metadata(line, "Hints").or(validate_hints);
-
-        fn num(s: &mut Scanner) -> Option<isize> {
-            let mut first = true;
-            let n = &s.eat_while(|c: char| {
-                let valid = first && c == '-' || c.is_numeric();
-                first = false;
-                valid
-            });
-            n.parse().ok()
-        }
-
-        let comments_until_code =
-            lines[i..].iter().take_while(|line| line.starts_with("//")).count();
-
-        let pos = |s: &mut Scanner| -> Option<usize> {
-            let first = num(s)? - 1;
-            let (delta, column) =
-                if s.eat_if(':') { (first, num(s)? - 1) } else { (0, first) };
-            let line = (i + comments_until_code).checked_add_signed(delta)?;
-            source.line_column_to_byte(line, usize::try_from(column).ok()?)
-        };
-
-        let range = |s: &mut Scanner| -> Option<Range<usize>> {
-            let start = pos(s)?;
-            let end = if s.eat_if('-') { pos(s)? } else { start };
-            Some(start..end)
-        };
-
-        for kind in AnnotationKind::iter() {
-            let Some(expectation) = get_metadata(line, kind.as_str()) else { continue };
-            let mut s = Scanner::new(expectation);
-            let range = range(&mut s);
-            let rest = if range.is_some() { s.after() } else { s.string() };
-            let message = rest
-                .trim()
-                .replace("VERSION", &PackageVersion::compiler().to_string())
-                .into();
-            annotations.insert(Annotation { kind, range, message });
-        }
-    }
-
-    TestPartMetadata {
-        part_configuration: TestConfiguration { compare_ref, validate_hints },
-        annotations,
+        write!(output, "{start_line}:{start_col}-{end_line}:{end_col} ").unwrap();
     }
+    writeln!(output, "{text}").unwrap();
 }
 
 /// Pseudorandomly edit the source file and test whether a reparse produces the
@@ -894,19 +1039,19 @@ fn test_spans_impl(output: &mut String, node: &SyntaxNode, within: Range<u64>) -
 }
 
 /// Draw all frames into one image with padding in between.
-fn render(frames: &[Frame]) -> sk::Pixmap {
+fn render(document: &Document) -> sk::Pixmap {
     let pixel_per_pt = 2.0;
     let padding = Abs::pt(5.0);
 
-    for frame in frames {
+    for page in &document.pages {
         let limit = Abs::cm(100.0);
-        if frame.width() > limit || frame.height() > limit {
-            panic!("overlarge frame: {:?}", frame.size());
+        if page.frame.width() > limit || page.frame.height() > limit {
+            panic!("overlarge frame: {:?}", page.frame.size());
         }
     }
 
     let mut pixmap = typst_render::render_merged(
-        frames,
+        document,
         pixel_per_pt,
         Color::WHITE,
         padding,
@@ -915,11 +1060,12 @@ fn render(frames: &[Frame]) -> sk::Pixmap {
 
     let padding = (pixel_per_pt * padding.to_pt() as f32).round();
     let [x, mut y] = [padding; 2];
-    for frame in frames {
+    for page in &document.pages {
         let ts =
             sk::Transform::from_scale(pixel_per_pt, pixel_per_pt).post_translate(x, y);
-        render_links(&mut pixmap, ts, frame);
-        y += (pixel_per_pt * frame.height().to_pt() as f32).round().max(1.0) + padding;
+        render_links(&mut pixmap, ts, &page.frame);
+        y += (pixel_per_pt * page.frame.height().to_pt() as f32).round().max(1.0)
+            + padding;
     }
 
     pixmap
